import os
import time
import shutil
from glob import glob
from pathlib import Path
from copy import deepcopy

import numpy as np
import wx

import larch
from larch import Group, Journal, Entry
from larch.larchlib import read_config, save_config
from larch.utils import (group2dict, unique_name, fix_varname, get_cwd,
                         asfloat, get_sessionid, mkdir, unixpath)
from larch.wxlib.plotter import last_cursor_pos
from larch.wxlib import ExceptionPopup
from larch.io import save_session
from larch.site_config import home_dir, user_larchdir

from .config import XASCONF, CONF_FILE,  OLDCONF_FILE

class XASController():
    """
    class holding the Larch session and doing the processing work for Larix
    """
    def __init__(self, wxparent=None, _larch=None):
        self.wxparent = wxparent
        self.filelist = None
        self.group = None
        self.groupname = None
        self.plot_erange = None
        self.report_frame = None
        self.recentfiles = []
        self.panels = {}
        self.larch = _larch
        if _larch is None:
            self.larch = larch.Interpreter()
        self.larix_folder = Path(user_larchdir, 'larix').as_posix()
        self.config_file = Path(self.larix_folder, CONF_FILE).as_posix()
        self.init_larch_session()
        self.init_workdir()

    def init_larch_session(self):
        self.symtable = self.larch.symtable
        self.file_groups = self.symtable._xasgroups = {}

        config = {}
        config.update(XASCONF)
        # may migrate old 'xas_viewer' folder to 'larix' folder
        xasv_folder = Path(user_larchdir, 'xas_viewer')
        if Path(xasv_folder).exists() and not Path(self.larix_folder).exists():
            print("Migrating xas_viewer to larix folder")
            shutil.move(xasv_folder, self.larix_folder)

        if not Path(self.larix_folder).exists():
            try:
                mkdir(self.larix_folder)
            except:
                title = "Cannot create Larix folder"
                message = [f"Cannot create directory {larix_folder}"]
                ExceptionPopup(self, title, message)


        # may migrate old 'xas_viewer.conf' file to 'larix.conf'
        old_config_file = Path(self.larix_folder, OLDCONF_FILE).as_posix()
        if Path(old_config_file).exists() and not Path(self.config_file).exists():
            shutil.move(old_config_file, self.config_file)

        if Path(self.config_file).exists():
            user_config = read_config(self.config_file)
            if user_config is not None:
                for sname in config:
                    if sname in user_config:
                        val = user_config[sname]
                        if isinstance(val, dict):
                            for k, v in val.items():
                                config[sname][k] = v
                        else:
                            config[sname] = val

        self.config = self.larch.symtable._sys.larix_config = config
        self.larch.symtable._sys.wx.plotopts = config['plot']
        self.clean_autosave_sessions()


    def install_group(self, groupname, filename, source=None, journal=None):
        """add groupname / filename to list of available data groups"""

        try:
            thisgroup = getattr(self.symtable, groupname)
        except AttributeError:
            thisgroup = self.symtable.new_group(groupname)

        # file /group may already exist in list
        if filename in self.file_groups:
            fbase, i = filename, 0
            while i < 50000 and filename in self.file_groups:
                filename = f"{fbase}_{i}"
                i += 1
                if i >= 50000:
                    raise ValueError(f"Too many repeated filenames: {fbase}")

        filename = filename.strip()
        if source is None:
            source = filename

        jopts = f"source='{source}'"
        if isinstance(journal, dict):
            jnl =  {'source': f"{source}"}
            jnl.update(journal)
            jopts = ', '.join([f"{k}='{v}'" for k, v in jnl.items()])
        elif isinstance(journal, (list, Journal)):
            jopts = repr(journal)

        cmds = [f"{groupname:s}.groupname = '{groupname:s}'",
                f"{groupname:s}.filename = '{filename:s}'"]
        needs_config = not hasattr(thisgroup, 'config')
        if needs_config:
            cmds.append(f"{groupname:s}.config = group(__name__='larix config')")

        cmds.append(f"{groupname:s}.journal = journal({jopts:s})")

        if hasattr(thisgroup, 'xdat') and not hasattr(thisgroup, 'xplot'):
            thisgroup.xplot = deepcopy(thisgroup.xdat)
        if hasattr(thisgroup, 'ydat') and not hasattr(thisgroup, 'yplot'):
            thisgroup.yplot = deepcopy(thisgroup.ydat)

        datatype = getattr(thisgroup, 'datatype', 'xydata')
        if datatype == 'xas':
            cmds.append(f"{groupname:s}.energy_orig = {groupname:s}.energy[:]")
            array_labels = getattr(thisgroup, 'array_labels', [])
            if len(array_labels) > 2  and getattr(thisgroup, 'data', None) is not None:
                for i0name in ('i0', 'i_0', 'monitor'):
                    if i0name in array_labels:
                        i0x = array_labels.index(i0name)
                        cmds.append(f"{groupname:s}.i0 = {groupname:s}.data[{i0x}, :]")

        self.larch.eval('\n'.join(cmds))

        if needs_config:
            # print("INSTALL GROUP, needs config", thisgroup)
            # print('\n'.join(cmds))
            self.init_group_config(thisgroup)

        self.file_groups[filename] = groupname
        self.filelist.Append(filename)
        self.filelist.SetStringSelection(filename)
        self.sync_xasgroups()
        return filename

    def sync_xasgroups(self):
        "make sure `_xasgroups` is identical to file_groups"
        if self.file_groups != self.symtable._xasgroups:
            self.symtable._xasgroups = self.file_groups

    def get_config(self, key, default=None):
        "get top-level, program-wide configuration setting"
        if key not in self.config:
            return default
        return deepcopy(self.config[key])

    def init_group_config(self, dgroup):
        """set up 'config' group with values from self.config"""
        if not hasattr(dgroup, 'config'):
            dgroup.config = larch.Group(__name__='larix config')

        for sect in ('exafs', 'feffit', 'lincombo', 'pca', 'prepeaks',
                     'regression', 'xasnorm'):
            setattr(dgroup.config, sect, deepcopy(self.config[sect]))

    def get_plot_conf(self):
        """get basic plot options to pass to plot() ** not window sizes **"""
        dx = {'linewidth': 3, 'markersize': 4,
              'show_grid': True, 'show_fullbox': True, 'theme': 'light'}
        pconf = self.config['plot']
        out = {}
        for attr, val in dx.items():
            out[attr] = pconf.get(attr, val)
        return out

    def save_config(self):
        """save configuration"""
        save_config(self.config_file, self.config)

    def chdir_on_fileopen(self):
        return self.config['main']['chdir_on_fileopen']

    def set_workdir(self):
        self.config['main']['workdir'] = get_cwd()

    def save_workdir(self):
        """save last workdir and recent session files"""
        try:
            with open(Path(self.larix_folder, 'workdir.txt'), 'w') as fh:
                fh.write(f"{get_cwd()}\n")
        except:
            pass

        buffer = []
        rfiles = []
        for tstamp, fname in sorted(self.recentfiles, key=lambda x: x[0], reverse=True)[:10]:
            if fname not in rfiles:
                buffer.append(f"{tstamp:.1f} {fname:s}")
                rfiles.append(fname)
        buffer.append('')
        buffer = '\n'.join(buffer)

        try:
            with open(Path(self.larix_folder, 'recent_sessions.txt'), 'w') as fh:
                fh.write(buffer)
        except:
            pass

    def init_workdir(self):
        """set initial working folder, read recent session files"""
        if self.config['main'].get('use_last_workdir', False):
            wfile = Path(self.larix_folder, 'workdir.txt')
            if wfile.exists():
                try:
                    with open(wfile, 'r') as fh:
                        workdir = fh.readlines()[0][:-1]
                        self.config['main']['workdir'] = workdir
                except:
                    pass
            try:
                os.chdir(self.config['main']['workdir'])
            except:
                pass

        rfile = Path(self.larix_folder, 'recent_sessions.txt')
        if rfile.exists():
            with open(rfile, 'r') as fh:
                for line in fh.readlines():
                    if len(line) < 2 or line.startswith('#'):
                        continue
                    try:
                        w = line[:-1].split(None, maxsplit=1)
                        self.recentfiles.insert(0, (float(w[0]), w[1]))
                    except:
                        pass


    def autosave_session(self):
        conf = self.get_config('autosave', {})
        fileroot = conf.get('fileroot', 'autosave')
        nhistory = max(8, int(conf.get('nhistory', 4)))

        fname =  f"{fileroot:s}_{get_sessionid():s}.larix"
        savefile = Path(self.larix_folder, fname).as_posix()
        for i in reversed(range(1, nhistory)):
            curf = savefile.replace('.larix', f'_{i:d}.larix' )
            if Path(curf).exists():
                newf = savefile.replace('.larix', f'_{i+1:d}.larix' )
                shutil.move(curf, newf)
        if Path(savefile).exists():
            curf = savefile.replace('.larix', '_1.larix' )
            shutil.move(savefile, curf)
        save_session(savefile, _larch=self.larch)
        return savefile

    def clean_autosave_sessions(self):
        conf = self.get_config('autosave', {})
        fileroot = conf.get('fileroot', 'autosave')
        max_hist = int(conf.get('maxfiles', 10))

        def get_autosavefiles():
            dat = []
            for afile in os.listdir(self.larix_folder):
                ffile = Path(self.larix_folder, afile)
                if afile.endswith('.larix'):
                    mtime = os.stat(ffile).st_mtime
                    words = afile.replace('.larix', '').split('_')
                    try:
                        version = int(words[-1])
                        words.pop()
                    except:
                        version = 0
                    dat.append((ffile.as_posix(), version, mtime))
            return sorted(dat, key=lambda x: x[2])

        dat = get_autosavefiles()
        nremove = max(0, len(dat) - max_hist)
        # first remove oldest "version > 0" files
        while nremove > 0 and len(dat) > 0:
            dfile, version, mtime = dat.pop(0)
            if version > 0:
                os.unlink(dfile)
                nremove -= 1

        dat = get_autosavefiles()
        nremove = max(0, len(dat) - max_hist)
        # then remove the oldest "version 0" files

        while nremove > 0 and len(dat) > 0:
            dfile, vers, mtime = dat.pop(0)
            if vers == 0 and abs(mtime - time.time()) > 86400:
                os.unlink(dfile)
            nremove -= 1

    def get_recentfiles(self, max=10):
        return sorted(self.recentfiles, key=lambda x: x[0], reverse=True)[:max]

    def recent_autosave_sessions(self):
        "return list of (timestamp, name) for most recent autosave session files"
        conf = self.get_config('autosave', {})
        fileroot = conf.get('fileroot', 'autosave')
        max_hist = int(conf.get('maxfiles', 10))
        flist = []
        for afile in os.listdir(self.larix_folder):
            ffile = Path(self.larix_folder, afile).as_posix()
            if ffile.endswith('.larix'):
                flist.append((os.stat(ffile).st_mtime, ffile))

        return sorted(flist, key=lambda x: x[0], reverse=True)[:max_hist]


    def clear_session(self):
        self.larch.eval("clear_session()")
        self.filelist.Clear()
        self.init_larch_session()


    def write_message(self, msg, panel=0):
        """write a message to the Status Bar"""
        self.wxparent.statusbar.SetStatusText(msg, panel)

    def close_all_displays(self):
        "close all displays, as at exit"
        self.symtable._plotter.close_all_displays()

    def get_display(self, win=1, stacked=False,
                    size=None, position=None):
        wintitle='Larch XAS Plot Window %i' % win

        conf = self.get_config('plot')
        opts = dict(wintitle=wintitle, stacked=stacked,
                    size=size, position=position, win=win)
        opts.update(conf)
        return self.symtable._plotter.get_display(**opts)

    def set_focus(self, topwin=None):
        """
        set wx focus to main window or selected Window,
        even after plot
        """
        if topwin is None:
            topwin = wx.GetApp().GetTopWindow()
            flist = self.filelist
        else:
            flist = getattr(topwin, 'filelist', topwin)
        time.sleep(0.025)
        topwin.Raise()

    def get_group(self, groupname=None):
        if groupname is None:
            groupname = self.groupname
            if groupname is None:
                return None
        dgroup = getattr(self.symtable, groupname, None)
        if dgroup is None and groupname in self.file_groups:
            groupname = self.file_groups[groupname]
            dgroup = getattr(self.symtable, groupname, None)

        if dgroup is None and len(self.file_groups) > 0:
            gname = list(self.file_groups.keys())[0]
            dgroup = getattr(self.symtable, gname, None)
        return dgroup

    def filename2group(self, filename):
        "convert filename (as displayed) to larch group"
        return self.get_group(self.file_groups[str(filename)])

    def merge_groups(self, grouplist, master=None, yarray='mu', outgroup=None):
        """merge groups"""
        if master is None:
            master = grouplist[0]
        gmaster = self.get_group(master)
        xarray = 'xplot' if gmaster.datatype=='xydata' else 'energy'
        outgroup = fix_varname(outgroup.lower())
        if outgroup is None:
            outgroup = 'merged'
        outgroup = unique_name(outgroup, self.file_groups, max=1000)

        glist = "[%s]" % (', '.join(grouplist))

        cmd = f"""{outgroup} = merge_groups({glist}, master={master},
        xarray='{xarray}', yarray='{yarray}', kind='cubic', trim=True)"""
        self.larch.eval(cmd)

        this = self.get_group(outgroup)
        if not hasattr(gmaster, 'config'):
            self.init_group_config(gmaster)
        if not hasattr(this, 'config'):
            self.init_group_config(this)
        this.config.xasnorm.update(gmaster.config.xasnorm)
        this.datatype = gmaster.datatype
        if xarray == 'energy':
            this.xplot = 1.0*this.energy
        this.yplot = 1.0*getattr(this, yarray)
        this.yerr =  getattr(this, 'd' + yarray, 1.0)
        if yarray != 'mu':
            this.mu = this.yplot
        this.plot_xlabel = xarray
        this.plot_ylabel = yarray
        return this

    def set_plot_erange(self, erange):
        self.plot_erange = erange

    def copy_group(self, filename, new_filename=None):
        """copy XAS group (by filename) to new group"""
        groupname = self.file_groups[filename]
        if not hasattr(self.larch.symtable, groupname):
            return

        ogroup = self.get_group(groupname)
        ngroup = larch.Group(datatype=ogroup.datatype, copied_from=groupname)

        for attr in dir(ogroup):
            val = getattr(ogroup, attr, None)
            if val is not None:
                setattr(ngroup, attr, deepcopy(val))

        if new_filename is None:
            new_filename = filename + '_1'
        ngroup.filename = unique_name(new_filename, self.file_groups.keys())
        ngroup.groupname = unique_name(groupname, self.file_groups.values())
        ngroup.journal.add('source_desc', f"copied from '{filename:s}'")
        setattr(self.larch.symtable, ngroup.groupname, ngroup)
        return ngroup

    def get_cursor(self, win=None):
        """get last cursor from selected window"""
        return last_cursor_pos(win=win, _larch=self.larch)

    def plot_group(self, groupname=None, title=None, plot_yarrays=None,
                   new=True, **kws):
        ppanel = self.get_display(stacked=False).panel
        newplot = ppanel.plot
        oplot   = ppanel.oplot
        plotcmd = oplot
        viewlims = ppanel.get_viewlimits()
        if new:
            plotcmd = newplot

        dgroup = self.get_group(groupname)
        if not hasattr(dgroup, 'xplot'):
            if hasattr(dgroup, 'xdat'):
                dgroup.xplot = deepcopy(dgroup.xdat)
            else:
                print("Cannot plot group ", groupname)

        if ((getattr(dgroup, 'plot_yarrays', None) is None or
             getattr(dgroup, 'energy', None) is None or
             getattr(dgroup, 'mu', None) is None)):
            self.process(dgroup)

        if plot_yarrays is None and hasattr(dgroup, 'plot_yarrays'):
            plot_yarrays = dgroup.plot_yarrays

        popts = kws
        fname = Path(dgroup.filename).name
        if not 'label' in popts:
            popts['label'] = dgroup.plot_ylabel

        popts['xlabel'] = dgroup.plot_xlabel
        popts['ylabel'] = dgroup.plot_ylabel
        if getattr(dgroup, 'plot_y2label', None) is not None:
            popts['y2label'] = dgroup.plot_y2label

        plot_extras = None
        if new:
            if title is None:
                title = fname
            plot_extras = getattr(dgroup, 'plot_extras', None)

        popts['title'] = title

        narr = len(plot_yarrays) - 1
        for i, pydat in enumerate(plot_yarrays):
            yaname, yopts, yalabel = pydat
            popts.update(yopts)
            if yalabel is not None:
                popts['label'] = yalabel
            popts['delay_draw'] = (i != narr)

            plotcmd(dgroup.xplot, getattr(dgroup, yaname), **popts)
            plotcmd = oplot

        if plot_extras is not None:
            axes = ppanel.axes
            for etype, x, y, opts in plot_extras:
                if etype == 'marker':
                    popts = {'marker': 'o', 'markersize': 4,
                             'label': '_nolegend_',
                             'markerfacecolor': 'red',
                             'markeredgecolor': '#884444'}
                    popts.update(opts)
                    axes.plot([x], [y], **popts)
                elif etype == 'vline':
                    popts = {'ymin': 0, 'ymax': 1.0,
                             'color': '#888888'}
                    popts.update(opts)
                    axes.axvline(x, **popts)
        ppanel.canvas.draw()
        self.set_focus()



import time
import sys
import locale
from  threading import Thread
from pathlib import Path
import wx
import wx.adv, wx.richtext
from  wx.lib.mixins.inspection import InspectionMixin

from larch.site_config import icondir

LARIX_TITLE = "Larix: XAS Visualization and Analysis"
ICON_FILE = 'onecone.ico'
SPLASH_STYLE = wx.adv.SPLASH_CENTRE_ON_SCREEN|wx.adv.SPLASH_TIMEOUT

class LarixApp(wx.App, InspectionMixin):
    def __init__(self, _larch=None, filename=None, check_version=True,
                 mode='xas', with_c_locale=True, with_wx_inspect=False, **kws):
        self._larch = _larch
        self.filename = filename
        self.mode = mode
        self.with_c_locale = with_c_locale
        self.with_wx_inspect = with_wx_inspect
        self.check_version = check_version
        wx.App.__init__(self,**kws)

    def OnInit(self):
        wx.SystemOptions.SetOption("mac.window-plain-transition", 1)
        LarixSplashScreen(filename=self.filename,
                          check_version=self.check_version,
                          mode=self.mode,
                          with_wx_inspect=self.with_wx_inspect)
        return True

    def createApp(self):
        return True

    def InitLocale(self):
        """over-ride wxPython default initial locale"""
        if self.with_c_locale:
            self._initial_locale = None
            locale.setlocale(locale.LC_ALL, 'C')
        else:
            lang, enc = locale.getdefaultlocale()
            self._initial_locale = wx.Locale(lang, lang[:2], lang)
            locale.setlocale(locale.LC_ALL, lang)

class LarixSplashScreen(wx.adv.SplashScreen):
    def __init__(self, **kws):
        self.kws = kws
        bmp = wx.Image(Path(icondir, ICON_FILE).as_posix()).ConvertToBitmap()
        wx.adv.SplashScreen.__init__(self, bmp, SPLASH_STYLE, 9000, None, -1)
        self.import_thread = Thread(target=self.importer)
        self.import_thread.start()
        wx.CallAfter(self.ShowMain)

    def importer(self, evt=None):
        from larch.wxxas.xasgui import LarixFrame

    def ShowMain(self):
        self.import_thread.join()
        from larch.wxxas.xasgui import LarixFrame
        self.frame = LarixFrame(**self.kws)
        wx.GetApp().SetTopWindow(self.frame)
        return True



#!/usr/bin/env python
"""
Linear Combination panel
"""
import os
import sys
import time

import wx
import wx.lib.scrolledpanel as scrolled
import wx.dataview as dv
import numpy as np

from functools import partial

import lmfit
from lmfit.printfuncs import fit_report

from larch import Group
from larch.math import index_of
from larch.xafs import etok, ktoe

from larch.wxlib import (BitmapButton, FloatCtrl, FloatSpin, ToggleButton,
                         GridPanel, get_icon, SimpleText, pack, Button,
                         HLine, Choice, Check, CEN, LEFT, Font, FONTSIZE,
                         FONTSIZE_FW, MenuItem, FRAMESTYLE, COLORS,
                         set_color, FileSave, EditableListBox,
                         DataTableGrid)

from .taskpanel import TaskPanel
from .config import ARRAYS, Linear_ArrayChoices, PlotWindowChoices
from larch.io import write_ascii
from larch.utils import gformat

np.seterr(all='ignore')

# plot options:
Plot_Choices = ['Data + Sum', 'Data + Sum + Components']

DVSTYLE = dv.DV_SINGLE|dv.DV_VERT_RULES|dv.DV_ROW_LINES

MAX_COMPONENTS = 12

def make_lcfplot(dgroup, form, with_fit=True, nfit=0):
    """make larch plot commands to plot LCF fit from form"""
    form['group'] = dgroup.groupname
    form['filename'] = dgroup.filename
    form['nfit'] = nfit
    form['label'] = label = 'Fit #%2.2d' % (nfit+1)

    if 'win' not in form: form['win'] = 1
    kspace = form['arrayname'].startswith('chi')
    if kspace:
        kw = 0
        if len(form['arrayname']) > 3:
            kw = int(form['arrayname'][3:])
        form['plotopt'] = 'kweight=%d' % kw

        cmds = ["""plot_chik({group:s}, {plotopt:s}, delay_draw=False, label='data',
         show_window=False, title='{filename:s}, {label:s}', win={win:d})"""]

    else:
        form['plotopt'] = 'show_norm=False'
        if form['arrayname'] == 'norm':
            form['plotopt'] = 'show_norm=True'
        elif form['arrayname'] == 'flat':
            form['plotopt'] = 'show_flat=True'
        elif form['arrayname'] == 'dmude':
            form['plotopt'] = 'show_deriv=True'

        erange = form['ehi'] - form['elo']
        form['pemin'] = 10*int( (form['elo'] - 5 - erange/4.0) / 10.0)
        form['pemax'] = 10*int( (form['ehi'] + 5 + erange/4.0) / 10.0)

        cmds = ["""plot_mu({group:s}, {plotopt:s}, delay_draw=True, label='data',
        emin={pemin:.1f}, emax={pemax:.1f}, title='{filename:s}, {label:s}', win={win:d})"""]

    if with_fit and hasattr(dgroup, 'lcf_result'):
        with_comps = True # "Components" in form['plotchoice']
        delay = 'delay_draw=True' if with_comps else 'delay_draw=False'
        xarr = "{group:s}.lcf_result[{nfit:d}].xdata"
        yfit = "{group:s}.lcf_result[{nfit:d}].yfit"
        ycmp = "{group:s}.lcf_result[{nfit:d}].ycomps"
        cmds.append("plot(%s, %s, label='%s', zorder=30, %s, win={win:d})" % (xarr, yfit, label, delay))
        ncomps = len(dgroup.lcf_result[nfit].ycomps)
        if with_comps:
            for i, key in enumerate(dgroup.lcf_result[nfit].ycomps):
                delay = 'delay_draw=False' if i==(ncomps-1) else 'delay_draw=True'
                cmds.append("plot(%s, %s['%s'], label='%s', %s, win={win:d})" % (xarr, ycmp, key, key, delay))

    # if form['show_e0']:
    #     cmds.append("plot_axvline({e0:1f}, color='#DDDDCC', zorder=-10)")
    if form['show_fitrange']:
        cmds.append("plot_axvline({elo:1f}, color='#EECCCC', zorder=-10, win={win:d})")
        cmds.append("plot_axvline({ehi:1f}, color='#EECCCC', zorder=-10, win={win:d})")

    script = "\n".join(cmds)
    return script.format(**form)

class LinComboResultFrame(wx.Frame):
    def __init__(self, parent=None,  datagroup=None, mainpanel=None, **kws):
        wx.Frame.__init__(self, None, -1, title='Linear Combination Results',
                          style=FRAMESTYLE, size=(925, 675), **kws)
        self.parent = parent
        self.mainpanel = mainpanel
        self.datagroup = datagroup
        self.datasets = {}
        self.form = self.mainpanel.read_form()
        self.larch_eval = self.mainpanel.larch_eval
        self.current_fit = 0
        self.createMenus()
        self.build()

        if self.mainpanel is not None:
            symtab = self.mainpanel.larch.symtable
            xasgroups = getattr(symtab, '_xasgroups', None)
            if xasgroups is not None:
                for dname, dgroup in xasgroups.items():
                    dgroup = getattr(symtab, dgroup, None)
                    hist  = getattr(dgroup, 'lcf_result', None)
                    if hist is not None:
                        self.add_results(dgroup, show=False)

    def createMenus(self):
        self.menubar = wx.MenuBar()
        fmenu = wx.Menu()
        m = {}

        MenuItem(self, fmenu, "Export current fit as group",
                 "Export current fit to a new group in the main panel",  self.onExportGroupFit)

        MenuItem(self, fmenu, "Save Fit And Components for Current Group",
                 "Save Fit and Components to Data File for Current Group",  self.onSaveGroupFit)

        MenuItem(self, fmenu, "Save Statistics for Best N Fits for Current Group",
                 "Save Statistics and Weights for Best N Fits for Current Group",  self.onSaveGroupStats)

        MenuItem(self, fmenu, "Save Data and Best N Fits for Current Group",
                 "Save Data and Best N Fits for Current Group",  self.onSaveGroupMultiFits)

        fmenu.AppendSeparator()
        MenuItem(self, fmenu, "Save Statistics Report for All Fitted Groups",
                 "Save Statistics for All Fitted Groups",  self.onSaveAllStats)

        self.menubar.Append(fmenu, "&File")
        self.SetMenuBar(self.menubar)

    def build(self):
        sizer = wx.GridBagSizer(3, 3)
        sizer.SetVGap(3)
        sizer.SetHGap(3)

        splitter = wx.SplitterWindow(self, style=wx.SP_LIVE_UPDATE)
        splitter.SetMinimumPaneSize(200)

        dl = self.filelist = EditableListBox(splitter, self.ShowDataSet,
                                           size=(250, -1))
        set_color(self.filelist, 'list_fg', bg='list_bg')


        panel = scrolled.ScrolledPanel(splitter)

        self.SetMinSize((650, 600))

        self.font_fixedwidth = wx.Font(FONTSIZE_FW, wx.MODERN, wx.NORMAL, wx.NORMAL)

        self.wids = wids = {}
        wids['plot_one'] = Button(panel, 'Plot This Fit', size=(125, -1),
                                  action=self.onPlotOne)
        wids['plot_sel'] = Button(panel, 'Plot N Best Fits', size=(125, -1),
                                  action=self.onPlotSel)

        wids['plot_win'] = Choice(panel, choices=PlotWindowChoices,
                                  action=self.onPlotOne, size=(60, -1))
        wids['plot_win'].SetStringSelection('1')

        wids['plot_wtitle'] = SimpleText(panel, 'Plot Window: ')
        wids['plot_ntitle'] = SimpleText(panel, 'N fits to plot: ')

        wids['plot_nchoice'] = Choice(panel, size=(60, -1),
                                      choices=['%d' % i for i in range(1, 21)])
        wids['plot_nchoice'].SetStringSelection('5')

        wids['data_title'] = SimpleText(panel, 'Linear Combination Result: <> ',
                                        font=Font(FONTSIZE+2),
                                        size=(400, -1),
                                        colour=COLORS['title'], style=LEFT)
        wids['nfits_title'] = SimpleText(panel, 'showing 5 best fits')
        wids['fitspace_title'] = SimpleText(panel, 'Array Fit: ')

        copts = dict(size=(125, 30), default=True, action=self.onPlotOne)
        # wids['show_e0'] = Check(panel, label='show E0?', **copts)
        wids['show_fitrange'] = Check(panel, label='show fit range?', **copts)

        irow = 0
        sizer.Add(wids['data_title'], (irow, 0), (1, 3), LEFT)

        irow += 1
        sizer.Add(wids['nfits_title'],     (irow, 0), (1, 1), LEFT)
        sizer.Add(wids['fitspace_title'],  (irow, 1), (1, 2), LEFT)


        irow += 1
        self.wids['paramstitle'] = SimpleText(panel, '[[Parameters]]',
                                              font=Font(FONTSIZE+2),
                                              colour=COLORS['title'], style=LEFT)
        sizer.Add(self.wids['paramstitle'], (irow, 0), (1, 3), LEFT)


        pview = self.wids['params'] = dv.DataViewListCtrl(panel, style=DVSTYLE)
        pview.SetFont(self.font_fixedwidth)
        pview.SetMinSize((500, 200))
        pview.AppendTextColumn(' Parameter ', width=180)
        pview.AppendTextColumn(' Best-Fit Value', width=150)
        pview.AppendTextColumn(' Standard Error ', width=150)
        for col in range(3):
            this = pview.Columns[col]
            isort, align = True, wx.ALIGN_RIGHT
            if col == 0:
                align = wx.ALIGN_LEFT
            this.Sortable = isort
            this.Alignment = this.Renderer.Alignment = align

        irow += 1
        sizer.Add(self.wids['params'],       (irow,   0), (7, 2), LEFT)
        sizer.Add(self.wids['plot_one'],     (irow,   2), (1, 2), LEFT)

        sizer.Add(self.wids['plot_wtitle'],  (irow+1, 2), (1, 1), LEFT)
        sizer.Add(self.wids['plot_win'],     (irow+1, 3), (1, 1), LEFT)


        sizer.Add(self.wids['show_fitrange'],(irow+2, 2), (1, 2), LEFT)
        sizer.Add((5, 5),                    (irow+3, 2), (1, 2), LEFT)
        sizer.Add(self.wids['plot_sel'],     (irow+4, 2), (1, 2), LEFT)
        sizer.Add(self.wids['plot_ntitle'],  (irow+5, 2), (1, 1), LEFT)
        sizer.Add(self.wids['plot_nchoice'], (irow+5, 3), (1, 1), LEFT)
        # sizer.Add(self.wids['show_e0'],      (irow+3, 1), (1, 2), LEFT)


        irow += 7
        sizer.Add(HLine(panel, size=(675, 3)), (irow, 0), (1, 4), LEFT)

        sview = self.wids['stats'] = dv.DataViewListCtrl(panel, style=DVSTYLE)
        sview.SetFont(self.font_fixedwidth)
        sview.Bind(dv.EVT_DATAVIEW_SELECTION_CHANGED, self.onSelectFitStat)
        sview.AppendTextColumn(' Fit #', width=65)
        sview.AppendTextColumn(' N_vary', width=80)
        sview.AppendTextColumn(' N_eval', width=80)
        sview.AppendTextColumn(' \u03c7\u00B2', width=100)
        sview.AppendTextColumn(' \u03c7\u00B2_reduced', width=100)
        sview.AppendTextColumn(' R Factor', width=100)
        sview.AppendTextColumn(' Akaike Info', width=100)

        for col in range(sview.ColumnCount):
            this = sview.Columns[col]
            isort, align = True, wx.ALIGN_RIGHT
            if col == 0:
                align = wx.ALIGN_CENTER
            this.Sortable = isort
            this.Alignment = this.Renderer.Alignment = align

        sview.SetMinSize((700, 175))

        irow += 1
        title = SimpleText(panel, '[[Fit Statistics]]', font=Font(FONTSIZE+2),
                           colour=COLORS['title'], style=LEFT)
        sizer.Add(title, (irow, 0), (1, 4), LEFT)

        irow += 1
        sizer.Add(sview, (irow, 0), (1, 4), LEFT)

        irow += 1
        sizer.Add(HLine(panel, size=(675, 3)), (irow, 0), (1, 4), LEFT)

        irow += 1
        title = SimpleText(panel, '[[Weights]]', font=Font(FONTSIZE+2),
                           colour=COLORS['title'], style=LEFT)
        sizer.Add(title, (irow, 0), (1, 4), LEFT)
        self.wids['weightspanel'] = ppan = wx.Panel(panel)

        p1 = SimpleText(ppan, ' < Weights > ')
        os = wx.BoxSizer(wx.VERTICAL)
        os.Add(p1, 1, 3)
        pack(ppan, os)
        ppan.SetMinSize((700, 175))

        irow += 1
        sizer.Add(ppan, (irow, 0), (1, 4), LEFT)

        irow += 1
        sizer.Add(HLine(panel, size=(675, 3)), (irow, 0), (1, 4), LEFT)

        pack(panel, sizer)
        panel.SetupScrolling()

        splitter.SplitVertically(self.filelist, panel, 1)

        mainsizer = wx.BoxSizer(wx.VERTICAL)
        mainsizer.Add(splitter, 1, wx.GROW|wx.ALL, 5)

        pack(self, mainsizer)
        # self.SetSize((725, 750))
        self.Show()
        self.Raise()

    def ShowDataSet(self, evt=None):
        dataset = evt.GetString()
        group = self.datasets.get(evt.GetString(), None)
        if group is not None:
            self.show_results(datagroup=group)

    def add_results(self, dgroup, form=None, larch_eval=None, show=True):
        name = dgroup.filename
        if name not in self.filelist.GetItems():
            self.filelist.Append(name)
        self.datasets[name] = dgroup
        if show:
            self.show_results(datagroup=dgroup, form=form, larch_eval=larch_eval)

    def show_results(self, datagroup=None, form=None, larch_eval=None):
        if datagroup is not None:
            self.datagroup = datagroup
        if form is not None:
            self.form = form
        if larch_eval is not None:
            self.larch_eval = larch_eval

        form = self.form
        if form is None:
            form = self.mainpanel.read_form()
        datagroup = self.datagroup

        wids = self.wids
        wids['data_title'].SetLabel('Linear Combination Result: %s ' %  self.datagroup.filename)
        wids['show_fitrange'].SetValue(form['show_fitrange'])

        wids['stats'].DeleteAllItems()
        if not hasattr(self.datagroup, 'lcf_result'):
            return
        results = self.datagroup.lcf_result[:20]
        self.nresults = len(results)
        wids['nfits_title'].SetLabel('showing %i best results' % (self.nresults,))
        wids['fitspace_title'].SetLabel('Array Fit: %s' % ARRAYS.get(results[0].arrayname, 'unknown'))

        for i, res in enumerate(results):
            res.result.rfactor = getattr(res, 'rfactor', 0)
            args = ['%2.2d' % (i+1)]
            for attr in ('nvarys', 'nfev', 'chisqr', 'redchi', 'rfactor', 'aic'):
                val = getattr(res.result, attr)
                if isinstance(val, int):
                    val = '%d' % val
                elif attr in ('aic',):
                    val = "%.2f" % val
                else:
                    val = gformat(val, 10)
                args.append(val)
            wids['stats'].AppendItem(tuple(args))

        wpan = self.wids['weightspanel']
        wpan.DestroyChildren()

        wview = self.wids['weights'] = dv.DataViewListCtrl(wpan, style=DVSTYLE)
        wview.SetFont(self.font_fixedwidth)
        wview.Bind(dv.EVT_DATAVIEW_SELECTION_CHANGED, self.onSelectFitParam)
        wview.AppendTextColumn(' Fit #', width=65)
        wview.AppendTextColumn(' E shift', width=80)

        for i, cname in enumerate(form['comp_names']):
            wview.AppendTextColumn(cname, width=100)
        wview.AppendTextColumn('Total', width=100)

        for col in range(len(form['comp_names'])+2):
            this = wview.Columns[col]
            isort, align = True, wx.ALIGN_RIGHT
            if col == 0:
                align = wx.ALIGN_CENTER
            this.Sortable = isort
            this.Alignment = this.Renderer.Alignment = align

        for i, res in enumerate(results):
            args = ['%2.2d' % (i+1), "%.4f" % res.params['e0_shift'].value]
            for cname in form['comp_names'] + ['total']:
                val = '--'
                if cname in res.params:
                    val = "%.4f" % res.params[cname].value
                args.append(val)
            wview.AppendItem(tuple(args))

        os = wx.BoxSizer(wx.VERTICAL)
        os.Add(wview, 1, wx.GROW|wx.ALL)
        pack(wpan, os)

        wview.SetMinSize((700, 500))
        s1, s2 = self.GetSize()
        if s2 % 2 == 0:
            s2 = s2 + 1
        else:
            s2 = s2 - 1
        self.SetSize((s1, s2))
        self.show_fitresult(0)
        self.Refresh()

    def onSelectFitParam(self, evt=None):
        if self.wids['weights'] is None:
            return
        item = self.wids['weights'].GetSelectedRow()
        self.show_fitresult(item)

    def onSelectFitStat(self, evt=None):
        if self.wids['stats'] is None:
            return
        item = self.wids['stats'].GetSelectedRow()
        self.show_fitresult(item)

    def show_fitresult(self, n):
        fit_result = self.datagroup.lcf_result[n]
        self.current_fit = n
        wids = self.wids
        wids['nfits_title'].SetLabel('Showing Fit # %2.2d' % (n+1,))
        wids['fitspace_title'].SetLabel('Array Fit: %s' % ARRAYS.get(fit_result.arrayname, 'unknown'))
        wids['paramstitle'].SetLabel('[[Parameters for Fit # %2.2d]]' % (n+1))

        wids['params'].DeleteAllItems()

        for pname, par in fit_result.params.items():
            args = [pname, gformat(par.value, 10), '--']
            if par.stderr is not None:
                args[2] = gformat(par.stderr, 10)
            self.wids['params'].AppendItem(tuple(args))

    def onPlotOne(self, evt=None):
        self.form = self.mainpanel.read_form()
        self.form['show_fitrange'] = self.wids['show_fitrange'].GetValue()
        self.form['win'] = int(self.wids['plot_win'].GetStringSelection())
        self.larch_eval(make_lcfplot(self.datagroup,
                                     self.form, nfit=self.current_fit))
        self.parent.controller.set_focus(topwin=self)

    def onPlotSel(self, evt=None):
        if self.form is None or self.larch_eval is None:
            return
        self.form['show_fitrange'] = self.wids['show_fitrange'].GetValue()
        self.form['win'] = int(self.wids['plot_win'].GetStringSelection())
        form = self.form
        dgroup = self.datagroup

        form['plotopt'] = 'show_norm=True'
        if form['arrayname'] == 'dmude':
            form['plotopt'] = 'show_deriv=True'
        if form['arrayname'] == 'flat':
            form['plotopt'] = 'show_flat=True'

        erange = form['ehi'] - form['elo']
        form['pemin'] = 10*int( (form['elo'] - 5 - erange/4.0) / 10.0)
        form['pemax'] = 10*int( (form['ehi'] + 5 + erange/4.0) / 10.0)

        cmds = ["""plot_mu({group:s}, {plotopt:s}, delay_draw=True, label='data',
        emin={pemin:.1f}, emax={pemax:.1f}, title='{filename:s}', win={win:d})"""]

        nfits = int(self.wids['plot_nchoice'].GetStringSelection())
        for i in range(nfits):
            delay = 'delay_draw=True' if i<nfits-1 else 'delay_draw=False'
            xarr = "{group:s}.lcf_result[%i].xdata" % i
            yfit = "{group:s}.lcf_result[%i].yfit" % i
            lab = 'Fit #%2.2d' % (i+1)
            cmds.append("plot(%s, %s, label='%s', zorder=30, %s)" % (xarr, yfit, lab, delay))

        if form['show_fitrange']:
            cmds.append("plot_axvline({elo:1f}, color='#EECCCC', zorder=-10)")
            cmds.append("plot_axvline({ehi:1f}, color='#EECCCC', zorder=-10)")

        script = "\n".join(cmds)
        self.larch_eval(script.format(**form))
        self.parent.controller.set_focus(topwin=self)

    def onExportGroupFit(self, evt=None):
        "Export current fit to a new group in the main panel"

        nfit = self.current_fit
        dgroup = self.datagroup
        xarr = dgroup.lcf_result[nfit].xdata
        yfit = dgroup.lcf_result[nfit].yfit
        i0 = np.ones_like(xarr)

        controller = self.parent.controller
        label = f"lcf_fit_{nfit}"
        groupname = new_group = f"{dgroup.groupname}_{label}"
        filename = f"{dgroup.filename}_{label}"
        cmdstr = f"""{new_group} = group(name="{groupname}", groupname="{groupname}", filename="{filename}")"""
        controller.larch.eval(cmdstr)
        g = controller.symtable.get_group(new_group)
        g.energy = g.xplot = xarr
        g.mu = g.yplot = g.norm = yfit
        g.i0 = i0
        g.datatype = 'xas'
        controller.install_group(groupname, filename, source="exported from Linear Combo / Fit Results")

    def onSaveGroupFit(self, evt=None):
        "Save Fit and Compoents for current fit to Data File"
        nfit = self.current_fit
        dgroup = self.datagroup
        nfits = int(self.wids['plot_nchoice'].GetStringSelection())

        deffile = "%s_LinearFit%i.dat" % (dgroup.filename, nfit+1)
        wcards  = 'Data Files (*.dat)|*.dat|All files (*.*)|*.*'
        path = FileSave(self, 'Save Fit and Components to File',
                        default_file=deffile, wildcard=wcards)
        if path is None:
            return

        form  = self.form
        label = [' energy         ',
                 ' data           ',
                 ' best_fit       ']
        result = dgroup.lcf_result[nfit]

        header = ['Larch Linear Fit Result for Fit: #%2.2d' % (nfit+1),
                  'Dataset filename: %s ' % dgroup.filename,
                  'Larch group: %s ' % dgroup.groupname,
                  'Array name: %s' %  form['arrayname'],
                  'Energy fit range: [%f, %f]' % (form['elo'], form['ehi']),
                  'Components: ']
        for key, val in result.weights.items():
            header.append('  %s: %f' % (key, val))

        report = fit_report(result.result).split('\n')
        header.extend(report)

        out = [result.xdata, result.ydata, result.yfit]
        for compname, compdata in result.ycomps.items():
            label.append(' %s' % (compname + ' '*(max(1, 15-len(compname)))))
            out.append(compdata)

        label = ' '.join(label)
        write_ascii(path, header=header, label=label, *out)


    def onSaveGroupStats(self, evt=None):
        "Save Statistics and Weights for Best N Fits for the current group"
        dgroup = self.datagroup
        nfits = int(self.wids['plot_nchoice'].GetStringSelection())
        results = dgroup.lcf_result[:nfits]
        nresults = len(results)
        deffile = "%s_LinearStats%i.dat" % (dgroup.filename, nresults)
        wcards  = 'Data Files (*.dat)|*.dat|All files (*.*)|*.*'

        path = FileSave(self, 'Save Statistics and Weights for Best N Fits',
                        default_file=deffile, wildcard=wcards)
        if path is None:
            return
        form = self.form

        header = ['Larch Linear Fit Statistics for %2.2d best results' % (nresults),
                  'Dataset filename: %s ' % dgroup.filename,
                  'Larch group: %s ' % dgroup.groupname,
                  'Array name: %s' %  form['arrayname'],
                  'Energy fit range: [%f, %f]' % (form['elo'], form['ehi']),
                  'N_Data: %d' % len(results[0].xdata)]

        label = ['fit #', 'n_varys', 'n_eval', 'chi2',
                  'chi2_reduced', 'akaike_info', 'bayesian_info']
        label.extend(form['comp_names'])
        label.append('Total')
        for i in range(len(label)):
            if len(label[i]) < 13:
                label[i] = (" %s                " % label[i])[:13]
        label = ' '.join(label)

        out = []
        for i, res in enumerate(results):
            dat = [(i+1)]
            for attr in ('nvarys', 'nfev', 'chisqr', 'redchi', 'aic', 'bic'):
                dat.append(getattr(res.result, attr))
            for cname in form['comp_names'] + ['total']:
                val = 0.0
                if cname in res.params:
                    val = res.params[cname].value
                dat.append(val)
            out.append(dat)

        out = np.array(out).transpose()
        write_ascii(path, header=header, label=label, *out)

    def onSaveGroupMultiFits(self, evt=None):
        "Save Data and Best N Fits for the current group"
        dgroup = self.datagroup
        nfits = int(self.wids['plot_nchoice'].GetStringSelection())
        results = dgroup.lcf_result[:nfits]
        nresults = len(results)

        deffile = "%s_LinearFits%i.dat" % (dgroup.filename, nresults)
        wcards  = 'Data Files (*.dat)|*.dat|All files (*.*)|*.*'

        path = FileSave(self, 'Save Best N Fits',
                        default_file=deffile, wildcard=wcards)
        if path is None:
            return
        form = self.form
        header = ['Larch Linear Arrays for %2.2d best results' % (nresults),
                  'Dataset filename: %s ' % dgroup.filename,
                  'Larch group: %s ' % dgroup.groupname,
                  'Array name: %s' %  form['arrayname'],
                  'Energy fit range: [%f, %f]' % (form['elo'], form['ehi'])]

        label = [' energy         ',  ' data           ']
        label.extend([' fit_%2.2d         ' % i for i in range(nresults)])
        label = ' '.join(label)

        out = [results[0].xdata, results[0].ydata]
        for i, res in enumerate(results):
            out.append(results[i].yfit)

        write_ascii(path, header=header, label=label, *out)

    def onSaveAllStats(self, evt=None):
        "Save All Statistics and Weights "
        deffile = "LinearFitStats.csv"
        wcards  = 'CVS Files (*.csv)|*.csv|All files (*.*)|*.*'
        path = FileSave(self, 'Save Statistics Report',
                        default_file=deffile, wildcard=wcards)
        if path is None:
            return
        form = self.form

        out = ['# Larch Linear Fit Statistics Report (best results) %s' % time.ctime(),
               '# Array name: %s' %  form['arrayname'],
               '# Energy fit range: [%f, %f]' % (form['elo'], form['ehi'])]

        label = [('Data Set' + ' '*25)[:25],
                 'n_varys', 'chi-square',
                 'chi-square_red', 'akaike_info', 'bayesian_info']
        label.extend(form['comp_names'])
        label.append('Total')
        for i in range(len(label)):
            if len(label[i]) < 12:
                label[i] = (" %s                " % label[i])[:12]
        label = ', '.join(label)
        out.append('# %s' % label)

        for name, dgroup in self.datasets.items():
            res = dgroup.lcf_result[0]
            label = dgroup.filename
            if len(label) < 25:
                label = (label + ' '*25)[:25]
            dat = [label]
            for attr in ('nvarys', 'chisqr', 'redchi', 'aic', 'bic'):
                dat.append(gformat(getattr(res.result, attr), 10))
            for cname in form['comp_names'] + ['total']:
                val = 0
                if cname in res.params:
                    val = res.params[cname].value
                dat.append(gformat(val, 10))
            out.append(', '.join(dat))
        out.append('')

        with open(path, 'w', encoding=sys.getdefaultencoding()) as fh:
            fh.write('\n'.join(out))

class LinearComboPanel(TaskPanel):
    """Liear Combination Panel"""
    def __init__(self, parent, controller, **kws):
        TaskPanel.__init__(self, parent, controller, panel='lincombo', **kws)

    def process(self, dgroup, **kws):
        """ handle linear combo processing"""
        if self.skip_process:
            return
        form = self.read_form()
        conf = self.get_config(dgroup)
        for key in ('elo', 'ehi', 'max_ncomps', 'fitspace', 'all_combos',
                     'vary_e0', 'sum_to_one', 'show_fitrange'):
            conf[key]  = form[key]
        self.update_config(conf, dgroup=dgroup)

    def build_display(self):
        panel = self.panel
        wids = self.wids
        self.skip_process = True

        wids['fitspace'] = Choice(panel, choices=list(Linear_ArrayChoices.keys()),
                                 action=self.onFitSpace, size=(175, -1))
        wids['fitspace'].SetSelection(0)

        add_text = self.add_text

        opts = dict(digits=2, increment=1.0, relative_e0=False)
        defaults = self.get_defaultconfig()

        self.make_fit_xspace_widgets(elo=defaults['elo_rel'], ehi=defaults['ehi_rel'])

        wids['fit_group'] = Button(panel, 'Fit this Group', size=(150, -1),
                                   action=self.onFitOne)
        wids['fit_selected'] = Button(panel, 'Fit Selected Groups', size=(175, -1),
                                      action=self.onFitAll)

        wids['fit_group'].Disable()
        wids['fit_selected'].Disable()

        wids['show_results'] = Button(panel, 'Show Fit Results',
                                      action=self.onShowResults, size=(150, -1))
        wids['show_results'].Disable()

        wids['add_selected'] = Button(panel, 'Use Selected Groups as Components',
                                      size=(300, -1), action=self.onUseSelected)

        opts = dict(default=True, size=(75, -1), action=self.onPlotOne)

        wids['show_fitrange'] = Check(panel, label='show?', **opts)

        wids['vary_e0'] = Check(panel, label='Allow energy shift in fit?', default=False)
        wids['sum_to_one'] = Check(panel, label='Weights Must Sum to 1?', default=False)
        wids['all_combos'] = Check(panel, label='Fit All Combinations?', default=True)
        max_ncomps = self.add_floatspin('max_ncomps', value=5, digits=0, increment=1,
                                        min_val=0, max_val=MAX_COMPONENTS, size=(60, -1),
                                        with_pin=False)

        panel.Add(SimpleText(panel, 'Linear Combination Analysis',
                             size=(350, -1), **self.titleopts), style=LEFT, dcol=4)

        add_text('Array to Fit: ', newrow=True)
        panel.Add(wids['fitspace'], dcol=3)
        panel.Add(wids['show_results'])

        panel.Add(wids['fitspace_label'], newrow=True)
        panel.Add(self.elo_wids)
        add_text(' : ', newrow=False)
        panel.Add(self.ehi_wids)
        panel.Add(wids['show_fitrange'])

        panel.Add(HLine(panel, size=(625, 3)), dcol=5, newrow=True)

        add_text('Build Model : ')
        panel.Add(wids['add_selected'], dcol=4)

        collabels = [' File /Group Name   ', 'weight', 'min', 'max']
        colsizes = [325, 100, 100, 100]
        coltypes = ['str', 'float:12,4', 'float:12,4', 'float:12,4']
        coldefs  = ['', 1.0/MAX_COMPONENTS, 0.0, 1.0]

        self.font_fixedwidth = wx.Font(FONTSIZE_FW, wx.MODERN, wx.NORMAL, wx.NORMAL)
        wids['table'] = DataTableGrid(panel, nrows=MAX_COMPONENTS,
                                      collabels=collabels,
                                      datatypes=coltypes, defaults=coldefs,
                                      colsizes=colsizes)

        wids['table'].SetMinSize((700, 250))
        wids['table'].SetFont(self.font_fixedwidth)
        panel.Add(wids['table'], newrow=True, dcol=6)

        panel.Add(HLine(panel, size=(625, 3)), dcol=5, newrow=True)
        add_text('Fit with this Model: ')
        panel.Add(wids['fit_group'], dcol=2)
        panel.Add(wids['fit_selected'], dcol=3)
        add_text('Fit Options: ')
        panel.Add(wids['vary_e0'], dcol=2)
        panel.Add(wids['sum_to_one'], dcol=2)
        panel.Add((10, 10), dcol=1, newrow=True)
        panel.Add(wids['all_combos'], dcol=2)
        add_text('Max # Components: ', newrow=False)
        panel.Add(max_ncomps, dcol=2)

        panel.Add(HLine(panel, size=(625, 3)), dcol=5, newrow=True)
        # panel.Add(wids['saveconf'], dcol=4, newrow=True)
        panel.pack()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add((10, 10), 0, LEFT, 3)
        sizer.Add(panel, 1, LEFT, 3)
        pack(self, sizer)
        self.skip_process = False

    def onPanelExposed(self, **kws):
        # called when notebook is selected
        try:
            fname = self.controller.filelist.GetStringSelection()
            gname = self.controller.file_groups[fname]
            dgroup = self.controller.get_group(gname)
            self.ensure_xas_processed(dgroup)
            self.fill_form(dgroup)
        except:
            pass # print(" Cannot Fill prepeak panel from group ")

        lcf_result = getattr(self.larch.symtable, 'lcf_result', None)
        if lcf_result is  None:
            return
        self.wids['show_results'].Enable()
        self.skip_process = True
        selected_groups = []
        for r in lcf_result[:100]:
            for gname in r.weights:
                if gname not in selected_groups:
                    selected_groups.append(gname)

        if len(selected_groups) > 0:
            if len(selected_groups) >= MAX_COMPONENTS:
                selected_groups = selected_groups[:MAX_COMPONENTS]
            weight = 1.0/len(selected_groups)
            grid_data = []
            for grp in selected_groups:
                grid_data.append([grp, weight, 0, 1])

            self.wids['fit_group'].Enable()
            self.wids['fit_selected'].Enable()
            self.wids['table'].table.data = grid_data
            self.wids['table'].table.View.Refresh()
        self.skip_process = False


    def onFitSpace(self, evt=None):
        fitspace = self.wids['fitspace'].GetStringSelection()
        self.update_config(dict(fitspace=fitspace))

        arrname = Linear_ArrayChoices.get(fitspace, 'norm')
        self.update_fit_xspace(arrname)
        self.plot()

    def onComponent(self, evt=None, comp=None):
        if comp is None or evt is None:
            return

        comps = []
        for wname, wid in self.wids.items():
            if wname.startswith('compchoice'):
                pref, n = wname.split('_')
                if wid.GetSelection() > 0:
                    caomps.append((int(n), wid.GetStringSelection()))
                else:
                    self.wids["compval_%s" % n].SetValue(0)

        cnames = set([elem[1] for elem in comps])
        if len(cnames) < len(comps):
            comps.remove((comp, evt.GetString()))
            self.wids["compchoice_%2.2d" % comp].SetSelection(0)

        weight = 1.0 / len(comps)

        for n, cname in comps:
            self.wids["compval_%2.2d" % n].SetValue(weight)


    def fill_form(self, dgroup):
        """fill in form from a data group"""
        opts = self.get_config(dgroup, with_erange=True)
        self.dgroup = dgroup
        self.ensure_xas_processed(dgroup)
        defaults = self.get_defaultconfig()

        self.skip_process = True
        wids = self.wids

        for attr in ('all_combos', 'sum_to_one', 'show_fitrange'):
            wids[attr].SetValue(opts.get(attr, True))

        for attr in ('elo', 'ehi', ):
            val = opts.get(attr, None)
            if val is not None:
                wids[attr].SetValue(val)

        for attr in ('fitspace', ):
            if attr in opts:
                wids[attr].SetStringSelection(opts[attr])

        fitspace = self.wids['fitspace'].GetStringSelection()
        self.update_config(dict(fitspace=fitspace))
        arrname = Linear_ArrayChoices.get(fitspace, 'norm')
        self.update_fit_xspace(arrname)

        self.skip_process = False

    def read_form(self, dgroup=None):
        "read form, return dict of values"
        self.skiap_process = True
        if dgroup is None:
            dgroup = self.controller.get_group()
        self.dgroup = dgroup
        if dgroup is None:
            opts = {'group': '', 'filename': ''}
        else:
            opts = {'group': dgroup.groupname, 'filename':dgroup.filename}

        wids = self.wids
        for attr in ('elo', 'ehi', 'max_ncomps'):
            opts[attr] = wids[attr].GetValue()

        opts['fitspace'] = wids['fitspace'].GetStringSelection()

        for attr in ('all_combos', 'vary_e0', 'sum_to_one', 'show_fitrange'):
            opts[attr] = wids[attr].GetValue()

        for attr, wid in wids.items():
            if attr.startswith('compchoice'):
                opts[attr] = wid.GetStringSelection()
            elif attr.startswith('comp'):
                opts[attr] = wid.GetValue()

        comps, cnames, wval, wmin, wmax = [], [], [], [], []

        table_data = self.wids['table'].table.data
        for _cname, _wval, _wmin, _wmax in table_data:
            if _cname.strip() in ('', None) or len(_cname) < 1:
                break
            cnames.append(_cname)
            comps.append(self.controller.file_groups[_cname])
            wval.append("%.5f" % _wval)
            wmin.append("%.5f" % _wmin)
            wmax.append("%.5f" % _wmax)

        opts['comp_names'] = cnames
        opts['comps']   = ', '.join(comps)
        opts['weights'] = ', '.join(wval)
        opts['minvals'] = ', '.join(wmin)
        opts['maxvals'] = ', '.join(wmax)
        opts['func'] = 'lincombo_fit'
        if opts['all_combos']:
            opts['func'] = 'lincombo_fitall'

        opts['arrayname'] = Linear_ArrayChoices.get(opts['fitspace'], 'norm')
        self.skip_process = False
        return opts

    def onSaveConfigBtn(self, evt=None):
        conf = self.get_config()
        conf.update(self.read_form())
        self.set_defaultconfig(conf)

    def onUseSelected(self, event=None):
        """ use selected groups as standards"""
        self.skip_process = True
        selected_groups = self.controller.filelist.GetCheckedStrings()
        if len(selected_groups) == 0:
            return
        if len(selected_groups) >= MAX_COMPONENTS:
            selected_groups = selected_groups[:MAX_COMPONENTS]
        weight = 1.0/len(selected_groups)

        grid_data = []
        for grp in selected_groups:
            grid_data.append([grp, weight, 0, 1])

        self.wids['fit_group'].Enable()
        self.wids['fit_selected'].Enable()
        self.wids['table'].table.data = grid_data
        self.wids['table'].table.View.Refresh()
        self.skip_process = False

    def do_fit(self, groupname, form, plot=True):
        """run lincombo fit for a group"""
        form['gname'] = groupname
        dgroup = self.controller.get_group(groupname)
        self.ensure_xas_processed(dgroup)

        if len(groupname) == 0:
            print("no group to fit?")
            return

        script = """# do LCF for {gname:s}
lcf_result = {func:s}({gname:s}, [{comps:s}],
            xmin={elo:.4f}, xmax={ehi:.4f},
            arrayname='{arrayname:s}',
            sum_to_one={sum_to_one}, vary_e0={vary_e0},
            weights=[{weights:s}],
            minvals=[{minvals:s}],
            maxvals=[{maxvals:s}],
            max_ncomps={max_ncomps:.0f})
"""
        if form['all_combos']:
            script = "%s\n{gname:s}.lcf_result = lcf_result\n" % script
        else:
            script = "%s\n{gname:s}.lcf_result = [lcf_result]\n" % script

        self.larch_eval(script.format(**form))

        dgroup = self.controller.get_group(groupname)
        self.show_subframe('lcf_result',  LinComboResultFrame,
                           datagroup=dgroup, mainpanel=self)

        self.subframes['lcf_result'].add_results(dgroup, form=form,
                                                 larch_eval=self.larch_eval, show=plot)
        if plot:
            self.plot(dgroup=dgroup, with_fit=True)

    def onShowResults(self, event=None):
        self.show_subframe('lcf_result',  LinComboResultFrame, mainpanel=self)

    def onFitOne(self, event=None):
        """ handle process events"""
        if self.skip_process:
            return

        self.skip_process = True
        form = self.read_form()
        self.update_config(form)
        self.do_fit(form['group'], form)
        self.skip_process = False

    def onFitAll(self, event=None):
        """ handle process events"""
        if self.skip_process:
            return
        self.skip_process = True
        form = self.read_form()
        groups = self.controller.filelist.GetCheckedStrings()
        for i, sel in enumerate(groups):
            gname = self.controller.file_groups[sel]
            self.do_fit(gname, form, plot=(i==len(groups)-1))
        self.skip_process = False

    def plot(self, dgroup=None, with_fit=False):
        if self.skip_plotting:
            return

        if dgroup is None:
            dgroup = self.controller.get_group()

        form = self.read_form(dgroup=dgroup)
        script = make_lcfplot(dgroup, form, with_fit=with_fit, nfit=0)
        self.larch_eval(script)
        self.controller.set_focus()



#!/usr/bin/env python

import Ifeffit
import time

iff = Ifeffit.Ifeffit()

script = """
read_data(file = ../xafsdata/cu_metal_rt.xdi, label = 'energy i0 i1 xmu', group = cu)
spline(energy= cu.energy, xmu= cu.xmu, kweight=1,rbkg=0.8)
fftf(cu.chi, kmin=1, kmax=21, dk=2, kwindow='hanning', kweight=2, group=cu, rmax_out=20)
fftr(real=cu.chir_re, imag=cu.chir_im, rmin=0.1, rmax=15, dr=0)
"""

iff.ifeffit('show $&build')

n = 100

t0 = time.time()

for i in range(n):
    iff.ifeffit(script)

print( 'Ran %i processing steps in % .4f seconds' % (n, time.time() - t0))

iff.ifeffit("plot(cu.k, cu.chi*cu.k**2)")

# doplot = False
# ##- timer = debug##- timer()
# t0 = time.time()
# n = 10
# for i in range(n):
#     print ' --- ', i
#
#     cu = read_ascii('tests/data/cu_metal_rt.xdi')
#     # cu = read_ascii('tests/data/cu_%3.3i.dat' % i)
#     ##- timer.add('read data file')
#     autobk(cu.energy, cu.mutrans, group=cu, rbkg=0.80, debug=True)
#     ##- timer.add('autobk done')
#
#     xafsft(cu.k, cu.chi, kmin=1, kmax=21, dk=2, window='hanning',
#            kweight=2, group=cu, rmax_out=20)
#     ##- timer.add('xafsft done')
#
#     xafsift(cu.r, cu.chir, rmin=0.1, rmax=4, qmax_out=20,
#             dr=0, window='hanning', group=cu)
#     ##- timer.add('xafs ift done')
#
#     if doplot:
#         newplot(cu.energy, cu.mutrans)
#         plot(cu.energy, cu.bkg)
#         #  plot(cu.energy, cu.init_bkg)
#         #  plot(cu.spline_e, cu.spline_y, linewidth=0, marker='o', color='red')
#         #  plot(cu.spline_e, cu.spline_yinit, linewidth=0, marker='o', color='black')
#         ##- timer.add('plot e-space done')
#
#         newplot(cu.k, cu.k**2*cu.chi, win=2)
#         # plot(cu.q, cu.chiq_re, win=2)
#         # #   plot(cu.q, cu.chiq_im, color='black', style='dashed')
#         #
#         ##- timer.add('plot k-space done')
#
#         # winscale = int((1.05* max(cu.chir_mag)))
#         newplot(cu.r, cu.chir_mag, win=3, xmax=15)
#         # plot(cu.r, cu.rwin*winscale, win=3, ymax=int(winscale*1.05))
#
#         ##- timer.add('plot r-space done')
#     endif
#
#     #if i % 8  == 0:
#     #    ##- timer.show_report()
#     #endif
# endfor
#
# # ##- timer.show_report()
#
#
# # show(cu)
#
# print 'Ran %i processing steps in % .4f seconds' % (n, time.time() - t0)
# ;



#!/usr/bin/env python
"""
Base TestCase for asteval
"""
import ast
import os
import textwrap
import time
import unittest
from sys import version_info
from tempfile import NamedTemporaryFile


from io import StringIO

from larch import Interpreter

HAS_NUMPY = False
try:
    # noinspection PyUnresolvedReferences
    import numpy as np
    HAS_NUMPY = True
except ImportError:
    print("Warning: numpy not available... functionality will be limited.")
    pass


class TestCase(unittest.TestCase):
    """testing of asteval"""

    def setUp(self):
        self.interp = Interpreter()
        self.symtable = self.interp.symtable

        self.interp.eval('x = 1')
        self.outfile = '_stdout_'
        self.outfile = '_stdout_'

        self.interp.writer = NamedTemporaryFile('w', delete=False,
                                                prefix='larchtest')

        if not HAS_NUMPY:
            self.interp("arange = range")

    def read_stdout(self):
        time.sleep(0.1)
        self.interp.writer.flush()
        out = None
        with open(self.interp.writer.name) as inp:
            out = inp.read()
        return out

    def tearDown(self):
        pass


    # noinspection PyUnresolvedReferences
    def isvalue(self, sym, val):
        """assert that a symboltable symbol has a particular value"""
        tval = self.interp.symtable.get_symbol(sym)
        if isinstance(val, np.ndarray):
            return self.assertTrue(np.all(tval == val))
        else:
            return self.assertTrue(tval == val)

    def isnear(self, expr, val, places=7):
        """assert that a symboltable symbol is near a particular value"""
        oval = self.interp(expr)
        if HAS_NUMPY and isinstance(val, np.ndarray):
            for x, y in zip(oval, val):
                self.assertAlmostEqual(x, y, places=places)
        else:
            return self.assertAlmostEqual(oval, val, places=places)

    # noinspection PyUnresolvedReferences
    def istrue(self, expr):
        """assert that an expression evaluates to True"""
        val = self.interp(expr)
        if HAS_NUMPY and isinstance(val, np.ndarray):
            val = np.all(val)
        return self.assertTrue(val)

    # noinspection PyUnresolvedReferences
    def isfalse(self, expr):
        """assert that an expression evaluates to False"""
        val = self.interp(expr)
        if HAS_NUMPY and isinstance(val, np.ndarray):
            val = np.all(val)
        return self.assertFalse(val)

    def check_output(self, chk_str, exact=False):
        self.interp.writer.flush()
        out = self.read_stdout().split('\n')
        if out:
            if exact:
                return chk_str == out[0]
            return chk_str in out[0]
        return False

    def check_error(self, chk_type, chk_msg=''):
        #trye
        if chk_type is None:
            self.assertEqual(0, len(self.interp.error))
        else:
            errtype, errmsg = self.interp.error[0].get_error()
            self.assertEqual(errtype, chk_type)
            if chk_msg:
                self.assertTrue(chk_msg in errmsg)

class TestEval(TestCase):
    """testing of asteval"""

    def test_dict_index(self):
        """dictionary indexing"""
        self.interp("a_dict = {'a': 1, 'b': 2, 'c': 3, 'd': 4}")
        self.istrue("a_dict['a'] == 1")
        self.istrue("a_dict['d'] == 4")

    def test_list_index(self):
        """list indexing"""
        self.interp("a_list = ['a', 'b', 'c', 'd', 'o']")
        self.istrue("a_list[0] == 'a'")
        self.istrue("a_list[1] == 'b'")
        self.istrue("a_list[2] == 'c'")

    def test_tuple_index(self):
        """tuple indexing"""
        self.interp("a_tuple = (5, 'a', 'x')")
        self.istrue("a_tuple[0] == 5")
        self.istrue("a_tuple[2] == 'x'")

    def test_string_index(self):
        """string indexing"""
        self.interp("a_string = 'hello world'")
        self.istrue("a_string[0] == 'h'")
        self.istrue("a_string[6] == 'w'")
        self.istrue("a_string[-1] == 'd'")
        self.istrue("a_string[-2] == 'l'")

    def test_ndarray_index(self):
        """nd array indexing"""
        self.interp("a_ndarray = 5*arange(20)")
        self.istrue("a_ndarray[2] == 10")
        self.istrue("a_ndarray[4] == 20")

    def test_ndarrayslice(self):
        """array slicing"""
        self.interp("a_ndarray = arange(200).reshape(10, 20)")
        self.istrue("a_ndarray[1:3,5:7] == array([[25,26], [45,46]])")
        self.interp("y = arange(20).reshape(4, 5)")
        self.istrue("y[:,3]  == array([3, 8, 13, 18])")
        self.istrue("y[...,1]  == array([1, 6, 11, 16])")
        self.interp("y[...,1] = array([2, 2, 2, 2])")
        self.istrue("y[1,:] == array([5, 2, 7, 8, 9])")

    def test_while(self):
        """while loops"""
        self.interp(textwrap.dedent("""
            n=0
            while n < 8:
                n += 1
            #endwhile
            """))
        self.isvalue('n', 8)

        self.interp(textwrap.dedent("""
            n=0
            while n < 8:
                n += 1
                if n > 3:
                    break
                #endif
            else:
                n = -1
            #endwhile
            """))
        self.isvalue('n', 4)

        self.interp(textwrap.dedent("""
            n=0
            while n < 8:
                n += 1
            else:
                n = -1
            #endwhile
            """))
        self.isvalue('n', -1)

        self.interp(textwrap.dedent("""
            n, i = 0, 0
            while n < 10:
                n += 1
                if n % 2:
                    continue
                #endif
                i += 1
            #endwhile
            print( 'finish: n, i = ', n, i)
            """))
        self.isvalue('n', 10)
        self.isvalue('i', 5)

        self.interp(textwrap.dedent("""
            n=0
            while n < 10:
                n += 1
                print( ' n = ', n)
                if n > 5:
                    break
                #endif
            #endwhile
            print( 'finish: n = ', n)
            """))
        self.isvalue('n', 6)

    def test_while_continue(self):
        self.interp(textwrap.dedent("""
            n, i = 0, 0
            while n < 10:
                n += 1
                if n % 2:
                    continue
                #endif
                i += 1
            #endwhile
            print( 'finish: n, i = ', n, i)
            """))
        self.isvalue('n', 10)
        self.isvalue('i', 5)

    def test_while_break(self):
        self.interp(textwrap.dedent("""
            n = 0
            while n < 10:
                n += 1
                if n > 6:
                    break
                #endif
            #endwhile
            print( 'finish: n = ', n)
            """))
        self.isvalue('n', 7)

    # noinspection PyTypeChecker
    def test_assert(self):
        """test assert statements"""
        self.interp.error = []
        self.interp('n=6')
        self.interp('assert n==6')
        self.check_error(None)
        self.interp('assert n==7')
        self.check_error('AssertionError')

    def test_for(self):
        """for loops"""
        self.interp(textwrap.dedent("""
            n=0
            for i in arange(10):
                n += i
            #endfor
            """))
        self.isvalue('n', 45)

        self.interp(textwrap.dedent("""
            n=0
            for i in arange(10):
                n += i
            else:
                n = -1
            #endfor
            """))
        self.isvalue('n', -1)

    def test_for_break(self):
        self.interp(textwrap.dedent("""
            n=0
            for i in arange(10):
                n += i
                if n > 2:
                    break
                #endif
            else:
                n = -1
            #endfor
            """))
        self.isvalue('n', 3)

    def test_if(self):
        """runtime errors test"""
        self.interp(textwrap.dedent("""
            zero = 0
            if zero == 0:
                x = 1
             #endif
            if zero != 100:
                x = x+1
            #endif
            if zero > 2:
                x = x + 1
            else:
                y = 33
            #endif
            """))
        self.isvalue('x', 2)
        self.isvalue('y', 33)

    def test_print(self):
        """print (ints, str, ....)"""
        self.interp("print(31)")
        self.check_output('31\n', True)
        self.interp("print('%s = %.3f' % ('a', 1.2012345))")
        self.check_output('a = 1.201\n', True)
        self.interp("print('{0:s} = {1:.2f}'.format('a', 1.2012345))")
        self.check_output('a = 1.20\n', True)

    def test_repr(self):
        """repr of dict, list"""
        self.interp("x = {'a': 1, 'b': 2, 'c': 3}")
        self.interp("y = ['a', 'b', 'c']")
        self.interp("rep_x = repr(x['a'])")
        self.interp("rep_y = repr(y)")
        self.interp("rep_y , rep_x")
        self.interp("repr(None)")
        self.isvalue("rep_x", "1")
        self.isvalue("rep_y", "['a', 'b', 'c']")

    def test_cmp(self):
        """numeric comparisons"""
        self.istrue("3 == 3")
        self.istrue("3.0 == 3")
        self.istrue("3.0 == 3.0")
        self.istrue("3 != 4")
        self.istrue("3.0 != 4")
        self.istrue("3 >= 1")
        self.istrue("3 >= 3")
        self.istrue("3 <= 3")
        self.istrue("3 <= 5")
        self.istrue("3 < 5")
        self.istrue("5 > 3")
        self.isfalse("3 == 4")
        self.isfalse("3 > 5")
        self.isfalse("5 < 3")

    def test_bool(self):
        """boolean logic"""
        self.interp(textwrap.dedent("""
            yes = True
            no = False
            nottrue = False
            a = arange(7)"""))

        self.istrue("yes")
        self.isfalse("no")
        self.isfalse("nottrue")
        self.isfalse("yes and no or nottrue")
        self.isfalse("yes and (no or nottrue)")
        self.isfalse("(yes and no) or nottrue")
        self.istrue("yes or no and nottrue")
        self.istrue("yes or (no and nottrue)")
        self.isfalse("(yes or no) and nottrue")
        self.istrue("yes or not no")
        self.istrue("(yes or no)")
        self.isfalse("not (yes or yes)")
        self.isfalse("not (yes or no)")
        self.isfalse("not (no or yes)")
        self.istrue("not no or yes")
        self.isfalse("not yes")
        self.istrue("not no")

    def test_bool_coerce(self):
        """coercion to boolean"""
        self.istrue("1")
        self.isfalse("0")
        self.istrue("'1'")
        self.isfalse("''")
        self.istrue("[1]")
        self.isfalse("[]")
        self.istrue("(1)")
        self.istrue("(0,)")
        self.isfalse("()")
        self.istrue("dict(y=1)")
        self.isfalse("{}")

    # noinspection PyUnresolvedReferences
    def test_assignment(self):
        """variables assignment"""
        self.interp('n = 5')
        self.isvalue("n", 5)
        self.interp('s1 = "a string"')
        self.isvalue("s1", "a string")
        self.interp('b = (1,2,3)')
        self.isvalue("b", (1, 2, 3))
        if HAS_NUMPY:
            self.interp('a = 1.*arange(10)')
            self.isvalue("a", np.arange(10))
            self.interp('a[1:5] = 1 + 0.5 * arange(4)')
            self.isnear("a", np.array([0., 1., 1.5, 2., 2.5, 5., 6., 7., 8., 9.]))

    def test_names(self):
        """names test"""
        self.interp('nx = 1')
        self.interp('nx1 = 1')

    def test_syntaxerrors_1(self):
        """assignment syntax errors test"""
        for expr in ('class = 1', 'for = 1', 'if = 1', 'raise = 1',
                     '1x = 1', '1.x = 1', '1_x = 1'):
            self.interp(expr)
            self.check_error('SyntaxError')

    def test_unsupportednodes(self):
        """unsupported nodes"""
        for expr in ('f = lambda x: x*x', 'yield 10'):
            failed = False
            self.interp.eval(expr)
            self.check_error('NotImplementedError')

    def test_syntaxerrors_2(self):
        """syntax errors test"""
        for expr in ('x = (1/*)', 'x = 1.A', 'x = A.2'):

            self.interp(expr)
            self.check_error('SyntaxError')

    def test_runtimeerrors_1(self):
        """runtime errors test"""
        self.interp("zero = 0")
        self.interp("astr ='a string'")
        self.interp("atup = ('a', 'b', 11021)")
        self.interp("arr  = arange(20)")
        for expr, errname in (('x = 1/zero', 'ZeroDivisionError'),
                              ('x = zero + nonexistent', 'NameError'),
                              ('x = zero + astr', 'TypeError'),
                              ('x = zero()', 'TypeError'),
                              ('x = astr * atup', 'TypeError'),
                              ('x = arr.shapx', 'AttributeError'),
                              ('arr.shapx = 4', 'AttributeError'),
                              ('del arr.shapx', 'LookupError')):
            errtype, errmsg = None, None
            self.interp(expr)
            self.check_error(errname)

    # noinspection PyUnresolvedReferences
    def test_ndarrays(self):
        """simple ndarrays"""
        self.interp('n = array([11, 10, 9])')
        self.istrue("isinstance(n, ndarray)")
        self.istrue("len(n) == 3")
        self.isvalue("n", np.array([11, 10, 9]))
        self.interp('n = arange(20).reshape(5, 4)')
        self.istrue("isinstance(n, ndarray)")
        self.istrue("n.shape == (5, 4)")
        self.interp("myx = n.shape")
        self.interp("n.shape = (4, 5)")
        self.istrue("n.shape == (4, 5)")
        self.interp("a = arange(20)")
        self.interp("gg = a[1:13:3]")
        self.isvalue('gg', np.array([1, 4, 7, 10]))
        self.interp("gg[:2] = array([0,2])")
        self.isvalue('gg', np.array([0, 2, 7, 10]))
        self.interp('a, b, c, d = gg')
        self.isvalue('c', 7)
        self.istrue('(a, b, d) == (0, 2, 10)')

    def test_binop(self):
        """test binary ops"""
        self.interp('a = 10.0')
        self.interp('b = 6.0')
        self.istrue("a+b == 16.0")
        self.isnear("a-b", 4.0)
        self.istrue("a/(b-1) == 2.0")
        self.istrue("a*b     == 60.0")

    def test_unaryop(self):
        """test binary ops"""
        self.interp('a = -10.0')
        self.interp('b = -6.0')
        self.isnear("a", -10.0)
        self.isnear("b", -6.0)

    def otest_del(self):
        """test del function"""
        self.interp('a = -10.0')
        self.interp('b = -6.0')
        self.assertTrue('a' in self.symtable)
        self.assertTrue('b' in self.symtable)
        self.interp("del a")
        self.interp("del b")
        self.assertFalse('a' in self.symtable)
        self.assertFalse('b' in self.symtable)

    # noinspection PyUnresolvedReferences
    def test_math1(self):
        """builtin math functions"""
        self.interp('n = sqrt(4)')
        self.istrue('n == 2')
        self.isnear('sin(pi/2)', 1)
        self.isnear('cos(pi/2)', 0)
        self.istrue('exp(0) == 1')
        self.isnear('exp(1)', np.e)

    def test_list_comprehension(self):
        """test list comprehension"""
        self.interp('x = [i*i for i in range(4)]')
        self.isvalue('x', [0, 1, 4, 9])
        self.interp('x = [i*i for i in range(6) if i > 1]')
        self.isvalue('x', [4, 9, 16, 25])

    def test_ifexp(self):
        """test if expressions"""
        self.interp('x = 2')
        self.interp('y = 4 if x > 0 else -1')
        self.interp('z = 4 if x > 3 else -1')
        self.isvalue('y', 4)
        self.isvalue('z', -1)

    # noinspection PyUnresolvedReferences
    def test_index_assignment(self):
        """test indexing / subscripting on assignment"""
        self.interp('x = arange(10)')
        self.interp('l = [1,2,3,4,5]')
        self.interp('l[0] = 0')
        self.interp('l[3] = -1')
        self.isvalue('l', [0, 2, 3, -1, 5])
        self.interp('l[0:2] = [-1, -2]')
        self.isvalue('l', [-1, -2, 3, -1, 5])
        self.interp('x[1] = 99')
        self.isvalue('x', np.array([0, 99, 2, 3, 4, 5, 6, 7, 8, 9]))
        self.interp('x[0:2] = [9,-9]')
        self.isvalue('x', np.array([9, -9, 2, 3, 4, 5, 6, 7, 8, 9]))

    def test_reservedwords(self):
        """test reserved words"""
        for w in ('and', 'as', 'while', 'raise', 'else',
                  'class', 'del', 'def', 'import', 'None', 'True', 'False'):
            self.interp.error = []
            self.interp("%s=2" % w)
            self.check_error('SyntaxError')

    def test_raise(self):
        """test raise"""
        self.interp("raise NameError('bob')")
        self.check_error('NameError', 'bob')

    def test_tryexcept(self):
        """test try/except"""
        self.interp(textwrap.dedent("""
            x = 5
            try:
                x = x/0
            except ZeroDivisionError:
                print( 'Error Seen!')
                x = -999
            #endtry
            """))
        self.isvalue('x', -999)

        self.interp(textwrap.dedent("""
            x = -1
            try:
                x = x/0
            except ZeroDivisionError:
                pass
            #endtry
            """))
        self.isvalue('x', -1)

    def test_tryelsefinally(self):

        self.interp(textwrap.dedent("""
            def dotry(x, y):
                out, ok, clean = 0, False, False
                try:
                    out = x/y
                except ZeroDivisionError:
                    out = -1
                else:
                    ok = True
                finally:
                    clean = True
                #endtry
                print(' DO TRY ', out, ok, clean)
                return out, ok, clean
            #enddef

            """))
        self.interp("val, ok, clean = dotry(1.0, 2.0)")
        self.interp("print( val, ok, clean)")
        self.isnear("val", 0.5)
        self.isvalue("ok", True)
        self.isvalue("clean", True)

        self.interp("val, ok, clean = dotry(1, 0.0)")
        self.isvalue("val", -1)
        self.isvalue("ok", False)
        self.isvalue("clean", True)

    def test_function1(self):
        """test function definition and running"""
        self.interp(textwrap.dedent("""
            def fcn(x, scale=2):
                'test function'
                out = sqrt(x)
                if scale > 1:
                    out = out * scale
                #endif
                return out
            #enddef
            """))
        self.interp("a = fcn(4, scale=9)")
        self.isvalue("a", 18)
        self.interp("a = fcn(9, scale=0)")
        self.isvalue("a", 3)
        self.interp("print(fcn)")
        self.check_output('<Procedure fcn(x, scale=')
        self.interp("a = fcn()")
        self.check_error('TypeError', 'expected at least')
        self.interp("a = fcn(77.0, other='what?')")
        self.check_error('TypeError', 'extra keyword arguments')

    def test_function_vararg(self):
        """test function with var args"""
        self.interp(textwrap.dedent("""
            def fcn(*args):
                'test varargs function'
                out = 0
                for i in args:
                    out = out + i*i
                #endfor
                return out
            #enddef
            """))
        self.interp("o = fcn(1,2,3)")
        self.isvalue('o', 14)
        self.interp("print(fcn)")
        self.check_output('<Procedure fcn(')

    def test_function_kwargs(self):
        """test function with kw args, no **kws"""
        self.interp(textwrap.dedent("""
            def fcn(square=False, x=0, y=0, z=0, t=0):
                'test varargs function'
                out = 0
                for i in (x, y, z, t):
                    if square:
                        out = out + i*i
                    else:
                        out = out + i
                    #endif
                #endfor
                return out
            #enddef
            """))
        self.interp("print(fcn)")
        self.check_output('<Procedure fcn(square')
        self.interp("o = fcn(x=1, y=2, z=3, square=False)")
        self.isvalue('o', 6)
        self.interp("o = fcn(x=1, y=2, z=3, square=True)")
        self.isvalue('o', 14)
        self.interp("o = fcn(x=1, y=2, z=3, t=-2)")
        self.isvalue('o', 4)
        self.interp("o = fcn(x=1, y=2, z=3, t=-12, s=1)")
        self.check_error('TypeError', 'extra keyword arg')

    def test_function_kwargs1(self):
        """test function with **kws arg"""
        self.interp(textwrap.dedent("""
            def fcn(square=False, **kws):
                'test varargs function'
                out = 0
                for i in kws.values():
                    if square:
                        out = out + i*i
                    else:
                        out = out + i
                    #endif
                #endfor
                return out
            #enddef
            """))
        self.interp("print(fcn)")
        self.check_output('<Procedure fcn(square')
        self.interp("o = fcn(x=1, y=2, z=3, square=False)")
        self.isvalue('o', 6)
        self.interp("o = fcn(x=1, y=2, z=3, square=True)")
        self.isvalue('o', 14)

    def test_function_kwargs2(self):
        """test function with positional and **kws args"""
        self.interp(textwrap.dedent("""
            def fcn(x, y):
                'test function'
                return x + y**2
            #enddef
            """))
        self.interp("print(fcn)")
        self.check_output('<Procedure fcn(x,')
        self.interp("o = -1")
        self.interp("o = fcn(2, 1)")
        self.isvalue('o', 3)
        self.interp("o = fcn(x=1, y=2)")
        self.isvalue('o', 5)
        self.interp("o = fcn(y=2, x=7)")
        self.isvalue('o', 11)
        self.interp("o = fcn(1, y=2)")
        self.isvalue('o', 5)
        self.interp("o = fcn(1, x=2)")
        self.check_error('TypeError')

    def test_nested_functions(self):
        setup="""
        def a(x=10):
            if x > 5:
                return 1
            #endif
            return -1
        #enddef

        def b():
            return 2.5
        #enddef

        def c(x=10):
            x = a(x=x)
            y = b()
            return x + y
        #enddef
        """
        self.interp(textwrap.dedent(setup))
        self.interp("o1 = c()")
        self.interp("o2 = c(x=0)")
        self.isvalue('o1', 3.5)
        self.isvalue('o2', 1.5)


    def test_astdump(self):
        """test ast parsing and dumping"""
        astnode = self.interp.parse('x = 1')
        self.assertTrue(isinstance(astnode, ast.Module))
        self.assertTrue(isinstance(astnode.body[0], ast.Assign))
        self.assertTrue(isinstance(astnode.body[0].targets[0], ast.Name))
        self.assertTrue(isinstance(astnode.body[0].value, ast.Constant))
        self.assertTrue(astnode.body[0].targets[0].id == 'x')
        self.assertTrue(astnode.body[0].value.value == 1)
        dumped = self.interp.dump(astnode.body[0])
        self.assertTrue(dumped.startswith('Assign'))


    def test_exit_value(self):
        """test expression eval - last exp. is returned by interpreter"""
        z = self.interp("True")
        self.assertTrue(z)
        z = self.interp("x = 1\ny = 2\ny == x + x\n")
        self.assertTrue(z)
        z = self.interp("x = 42\nx")
        self.assertEqual(z, 42)
        self.isvalue('x', 42)
        z = self.interp("""def foo(): return 42\nfoo()""")
        self.assertEqual(z, 42)


if __name__ == '__main__':
    for suite in (TestEval,):
        suite = unittest.TestLoader().loadTestsFromTestCase(suite)
        unittest.TextTestRunner(verbosity=2).run(suite)



"""
A distutils script to make a standalone .exe of superdoodle for
Windows platforms.	You can get py2exe from
http://py2exe.sourceforge.net/.	 Use this command to build the .exe
and collect the other needed files:

    python setup.py py2exe

A distutils script to make a standalone .app of superdoodle for
Mac OS X.  You can get py2app from http://undefined.org/python/py2app.
Use this command to build the .app and collect the other needed files:

    python setup.py py2app
"""


from setuptools import setup
import sys
if sys.platform == 'darwin':
    buildstyle = 'app'
elif sys.platform == 'win32':
    import py2exe
    # buildstyle = 'console'
    buildstyle = 'windows'

setup(
    name="SimpleParratt_Panel",
    setup_requires=["py2app"],
    **{buildstyle : ['SimpleParratt_Panel.py']}
)





# -*- coding: utf-8 -*-
#
# larch documentation build configuration file, created by

import sys, os


CURDIR = os.path.abspath(os.path.dirname(__file__))

sys.path.insert(0, os.path.abspath(os.path.join('sphinx', 'ext')))
sys.path.insert(0, CURDIR)
sys.path.append(os.path.abspath(os.path.join('.')))

from extensions import extensions

import authorlist_format

intersphinx_mapping = {'py':    ('https://docs.python.org/3', None),
                       'numpy': ('https://numpy.org/doc/stable/', None),
                       'scipy': ('https://docs.scipy.org/doc/scipy/', None),
                       }

extlinks = {
    'scipydoc' : ('https://docs.scipy.org/doc/scipy/reference/generated/scipy.%s.html', 'scipy.'),
    'numpydoc' : ('https://docs.scipy.org/doc/numpy/reference/generated/numpy.%s.html', 'numpy.'),
    'lmfitdoc' : ('https://lmfit.github.io/lmfit-py/%s.html', 'lmfit.'),
    'lmfitx' : ('https://lmfit.github.io/lmfit-py/%s', ' '),
    }

# List of patterns, relative to source directory, that match files and
# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'xraylarch'
author = 'Matthew Newville'
copyright = 'Matthew Newville, The University of Chicago, 2022'

numfig = True
numfig_secnum_depth = 3
numfig_format = {'figure': 'Figure %s',
                 'table': 'Table %s',
                 'code-block': 'Listing %s',
                 'section': 'Section %s'}

bibtex_bibfiles = ['larch.bib']

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
try:
    import larch
    release = larch.__release_version__
# The full version, including alpha/beta/rc tags.
except ImportError:
    release = 'unknown (larch import failed??)'

exclude_trees = ['_build']
exclude_patterns = ['_build', 'sphinx', '_junk', 'epilog.rst']

#sphinxtr
# Ideally, we wouldn't have to do this, but sphinx seems to have trouble with
# directives inside only directives
if tags.has('latex'):
    master_doc = 'index_tex'
    exclude_patterns.append('index.rst')
else:
    master_doc = 'index'
    exclude_patterns.append('index_tex.rst')

#sphinxtr
# A string of reStructuredText that will be included at the end of
# every source file that is read.
rst_epilog = open(os.path.join(CURDIR, 'epilog.rst'),'r').read()# .decode('utf-8')

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = ['sphinx/theme']
# html_theme = 'nature'
html_theme = 'larchdoc'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

# html_theme_options = {'collapsiblesidebar': False}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title =

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typograpbhically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {'index': ['indexsidebar.html','searchbox.html']}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
html_use_modindex = False

htmlhelp_basename = 'larchdoc'
html_domain_indices = False



#!/usr/bin/env python
"""

"""
from pathlib import Path

import numpy as np
np.seterr(all='ignore')

from functools import partial

import wx
import wx.lib.scrolledpanel as scrolled
import wx.lib.agw.flatnotebook as fnb
from wxmplot import PlotPanel

from wxutils import (SimpleText, FloatCtrl, GUIColors, Button, Choice,
                     FileCheckList, pack, Popup, Check, MenuItem, CEN,
                     RIGHT, LEFT, FRAMESTYLE, HLine, Font)

import larch
from larch import Group
from larch.utils.strutils import fix_varname
from larch.xafs.xafsutils import guess_energy_units
from larch.io import look_for_nans, is_specfile, open_specfile
from larch.utils.physical_constants import PLANCK_HC, DEG2RAD, PI
from .columnframe import MultiColumnFrame, create_arrays, energy_may_need_rebinning

CEN |=  wx.ALL
FNB_STYLE = fnb.FNB_NO_X_BUTTON|fnb.FNB_SMART_TABS
FNB_STYLE |= fnb.FNB_NO_NAV_BUTTONS|fnb.FNB_NODRAG

XPRE_OPS = ('', 'log(', '-log(')
YPRE_OPS = ('', 'log(', '-log(')
ARR_OPS = ('+', '-', '*', '/')

YERR_OPS = ('Constant', 'Sqrt(Y)', 'Array')
CONV_OPS  = ('Lorenztian', 'Gaussian')

XDATATYPES = ('xydata', 'xas')
ENUNITS_TYPES = ('eV', 'keV', 'degrees', 'not energy')


class SpecfileImporter(wx.Frame) :
    """Column Data File, select columns"""
    def __init__(self, parent, filename=None, config=None, _larch=None,
                 read_ok_cb=None):
        if not is_specfile(filename):
            title = "Not a Specfile: %s" % filename
            message = "Error reading %s as a Specfile" % filename
            r = Popup(parent, message, title)
            return None

        self.parent = parent
        fpath = Path(filename).absolute()
        self.path = fpath.as_posix()
        self.filename = fpath.name
        self._larch = _larch
        self.specfile = open_specfile(self.path)
        self.scans = []
        curscan = None
        for scandata in self.specfile.get_scans():
            name, cmd, dtime = scandata
            self.scans.append("%s | %s" % (name, cmd))
            if curscan is None:
                curscan = name

        self.curscan = self.specfile.get_scan(curscan)
        self.subframes = {}
        self.workgroup = Group()
        for attr in ('path', 'filename', 'datatype',
                     'array_labels', 'data'):
            setattr(self.workgroup, attr, None)

        self.array_labels = [l.lower() for l in self.curscan.array_labels]

        if self.workgroup.datatype is None:
            self.workgroup.datatype = 'xydata'
            for arrlab in self.array_labels[:4]:
                if 'ener' in arrlab.lower():
                    self.workgroup.datatype = 'xas'

        self.read_ok_cb = read_ok_cb

        self.config = dict(xarr=self.curscan.axis.lower(), yarr1=None,
                           yarr2=None, yop='/', ypop='',
                           monod=3.1355316, en_units='eV',
                           yerror=YERR_OPS[0], yerr_val=1,
                           yerr_arr=None, dtc_config={}, multicol_config={})

        if config is not None:
            self.config.update(config)

        if self.config['yarr2'] is None and 'i0' in self.array_labels:
            self.config['yarr2'] = 'i0'

        if self.config['yarr1'] is None:
            if 'itrans' in self.array_labels:
                self.config['yarr1'] = 'itrans'
            elif 'i1' in self.array_labels:
                self.config['yarr1'] = 'i1'

        wx.Frame.__init__(self, None, -1, f'Build Arrays for {filename:s}',
                          style=FRAMESTYLE)

        self.SetMinSize((750, 550))
        self.SetSize((850, 650))
        self.colors = GUIColors()

        x0, y0 = parent.GetPosition()
        self.SetPosition((x0+60, y0+60))
        self.SetFont(Font(10))

        splitter  = wx.SplitterWindow(self, style=wx.SP_LIVE_UPDATE)
        splitter.SetMinimumPaneSize(200)

        leftpanel = wx.Panel(splitter)
        ltop = wx.Panel(leftpanel)

        sel_none = Button(ltop, 'Select None', size=(100, 30), action=self.onSelNone)
        sel_all  = Button(ltop, 'Select All', size=(100, 30), action=self.onSelAll)
        sel_imp  = Button(ltop, 'Import Selected Scans', size=(200, -1), action=self.onOK)

        self.scanlist = FileCheckList(leftpanel, select_action=self.onScanSelect)
        self.scanlist.AppendItems(self.scans)

        tsizer = wx.GridBagSizer(2, 2)
        tsizer.Add(sel_all, (0, 0), (1, 1), LEFT, 0)
        tsizer.Add(sel_none,  (0, 1), (1, 1), LEFT, 0)
        tsizer.Add(sel_imp,  (1, 0), (1, 2), LEFT, 0)
        pack(ltop, tsizer)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(ltop, 0, LEFT|wx.GROW, 1)
        sizer.Add(self.scanlist, 1, LEFT|wx.GROW|wx.ALL, 1)
        pack(leftpanel, sizer)

        # right hand side
        rightpanel = wx.Panel(splitter)

        panel = wx.Panel(rightpanel)
        # title row
        self.title = SimpleText(panel,
                                "  %s, scan %s" % (self.path, self.curscan.scan_name),
                                font=Font(11), colour=self.colors.title, style=LEFT)

        self.wid_scantitle = SimpleText(panel, " %s" % self.curscan.title,
                                       font=Font(11), style=LEFT)
        self.wid_scantime = SimpleText(panel, self.curscan.timestring,
                                       font=Font(11), style=LEFT)

        yarr_labels = self.yarr_labels = self.array_labels + ['1.0', '0.0', '']
        xarr_labels = self.xarr_labels = self.array_labels + ['_index']

        self.xarr   = Choice(panel, choices=xarr_labels, action=self.onXSelect, size=(150, -1))
        self.yarr1  = Choice(panel, choices=yarr_labels, action=self.onUpdate, size=(150, -1))
        self.yarr2  = Choice(panel, choices=yarr_labels, action=self.onUpdate, size=(150, -1))
        self.yerr_arr = Choice(panel, choices=yarr_labels, action=self.onUpdate, size=(150, -1))
        self.yerr_arr.Disable()

        self.datatype = Choice(panel, choices=XDATATYPES, action=self.onUpdate, size=(150, -1))
        self.datatype.SetStringSelection(self.workgroup.datatype)

        self.en_units = Choice(panel, choices=ENUNITS_TYPES, action=self.onEnUnitsSelect,
                               size=(150, -1))

        self.ypop = Choice(panel, choices=YPRE_OPS, action=self.onUpdate, size=(150, -1))
        self.yop =  Choice(panel, choices=ARR_OPS, action=self.onUpdate, size=(50, -1))
        self.yerr_op = Choice(panel, choices=YERR_OPS, action=self.onYerrChoice, size=(150, -1))

        self.yerr_val = FloatCtrl(panel, value=1, precision=4, size=(90, -1))
        self.monod_val  = FloatCtrl(panel, value=3.1355316, precision=7, size=(90, -1))

        xlab = SimpleText(panel, ' X array: ')
        ylab = SimpleText(panel, ' Y array: ')
        units_lab = SimpleText(panel, '  Units:  ')
        yerr_lab = SimpleText(panel, ' Yerror: ')
        dtype_lab = SimpleText(panel, ' Data Type: ')
        monod_lab = SimpleText(panel, ' Mono D spacing (Ang): ')
        yerrval_lab = SimpleText(panel, ' Value:')

        self.ysuf = SimpleText(panel, '')
        self.message = SimpleText(panel, '-', font=Font(11),
                           colour=self.colors.title, style=LEFT)

        self.ypop.SetStringSelection(self.config['ypop'])
        self.yop.SetStringSelection(self.config['yop'])
        self.monod_val.SetValue(self.config['monod'])
        self.monod_val.SetAction(self.onUpdate)

        self.monod_val.Enable(self.config['en_units'].startswith('deg'))
        self.en_units.SetStringSelection(self.config['en_units'])
        self.yerr_op.SetStringSelection(self.config['yerror'])
        self.yerr_val.SetValue(self.config['yerr_val'])
        if '(' in self.config['ypop']:
            self.ysuf.SetLabel(')')

        ixsel, iysel, iy2sel, iyesel = 0, 1, len(yarr_labels)-1,  len(yarr_labels)-1
        if self.config['xarr'] in xarr_labels:
            ixsel = xarr_labels.index(self.config['xarr'])
        if self.config['yarr1'] in self.array_labels:
            iysel = self.array_labels.index(self.config['yarr1'])
        if self.config['yarr2'] in yarr_labels:
            iy2sel = yarr_labels.index(self.config['yarr2'])
        if self.config['yerr_arr'] in yarr_labels:
            iyesel = yarr_labels.index(self.config['yerr_arr'])
        self.xarr.SetSelection(ixsel)
        self.yarr1.SetSelection(iysel)
        self.yarr2.SetSelection(iy2sel)
        self.yerr_arr.SetSelection(iyesel)

        bpanel = wx.Panel(panel)
        bsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.multi_sel = Button(bpanel, 'Select Multilple Columns',  action=self.onMultiColumn)
        self.multi_clear = Button(bpanel, 'Clear Multiple Columns',  action=self.onClearMultiColumn)
        self.multi_clear.Disable()
        self.multi_sel.SetToolTip('Select Multiple Columns to import as separate groups')
        self.multi_clear.SetToolTip('Clear Multiple Column Selection')
        bsizer.Add(self.multi_sel)
        bsizer.Add(self.multi_clear)
        pack(bpanel, bsizer)

        sizer = wx.GridBagSizer(2, 2)
        ir = 0
        sizer.Add(self.title,     (ir, 0), (1, 7), LEFT, 5)

        ir += 1
        sizer.Add(self.wid_scantitle,  (ir, 0), (1, 3), LEFT, 0)
        sizer.Add(self.wid_scantime,   (ir, 3), (1, 2), LEFT, 0)


        ir += 1
        sizer.Add(xlab,      (ir, 0), (1, 1), LEFT, 0)
        sizer.Add(self.xarr,  (ir, 1), (1, 1), LEFT, 0)
        sizer.Add(units_lab,     (ir, 2), (1, 2), RIGHT, 0)
        sizer.Add(self.en_units,  (ir, 4), (1, 2), LEFT, 0)

        ir += 1
        sizer.Add(dtype_lab,          (ir, 0), (1, 1), LEFT, 0)
        sizer.Add(self.datatype,      (ir, 1), (1, 1), LEFT, 0)
        sizer.Add(monod_lab,          (ir, 2), (1, 2), RIGHT, 0)
        sizer.Add(self.monod_val,     (ir, 4), (1, 1), LEFT, 0)

        ir += 1
        sizer.Add(ylab,       (ir, 0), (1, 1), LEFT, 0)
        sizer.Add(self.ypop,  (ir, 1), (1, 1), LEFT, 0)
        sizer.Add(self.yarr1, (ir, 2), (1, 1), LEFT, 0)
        sizer.Add(self.yop,   (ir, 3), (1, 1), RIGHT, 0)
        sizer.Add(self.yarr2, (ir, 4), (1, 1), LEFT, 0)
        sizer.Add(self.ysuf,  (ir, 5), (1, 1), LEFT, 0)

        ir += 1
        sizer.Add(yerr_lab,      (ir, 0), (1, 1), LEFT, 0)
        sizer.Add(self.yerr_op,  (ir, 1), (1, 1), LEFT, 0)
        sizer.Add(self.yerr_arr, (ir, 2), (1, 1), LEFT, 0)
        sizer.Add(yerrval_lab,   (ir, 3), (1, 1), RIGHT, 0)
        sizer.Add(self.yerr_val, (ir, 4), (1, 2), LEFT, 0)

        ir += 1
        sizer.Add(self.message,                     (ir, 0), (1, 4), LEFT, 0)
        ir +=1
        sizer.Add(bpanel,     (ir, 0), (1, 5), LEFT, 3)

        pack(panel, sizer)

        self.nb = fnb.FlatNotebook(rightpanel, -1, agwStyle=FNB_STYLE)
        self.nb.SetTabAreaColour(wx.Colour(248,248,240))
        self.nb.SetActiveTabColour(wx.Colour(254,254,195))
        self.nb.SetNonActiveTabTextColour(wx.Colour(40,40,180))
        self.nb.SetActiveTabTextColour(wx.Colour(80,0,0))

        self.plotpanel = PlotPanel(rightpanel, messenger=self.plot_messages)
        try:
            plotopts = self._larch.symtable._sys.wx.plotopts
            self.plotpanel.conf.set_theme(plotopts['theme'])
            self.plotpanel.conf.enable_grid(plotopts['show_grid'])
        except:
            pass

        self.plotpanel.SetMinSize((300, 250))

        shead = wx.Panel(rightpanel)
        self.scanheader = wx.TextCtrl(shead, style=wx.TE_MULTILINE|wx.TE_READONLY,
                                      size=(400, 250))
        self.scanheader.SetValue('\n'.join(self.curscan.scan_header))
        self.scanheader.SetFont(Font(10))
        textsizer = wx.BoxSizer(wx.VERTICAL)
        textsizer.Add(self.scanheader, 1, LEFT|wx.GROW, 1)
        pack(shead, textsizer)


        fhead = wx.Panel(rightpanel)
        self.fileheader = wx.TextCtrl(fhead, style=wx.TE_MULTILINE|wx.TE_READONLY,
                                      size=(400, 250))
        self.fileheader.SetValue('\n'.join(self.curscan.file_header))
        self.fileheader.SetFont(Font(10))
        textsizer = wx.BoxSizer(wx.VERTICAL)
        textsizer.Add(self.fileheader, 1, LEFT|wx.GROW, 1)
        pack(fhead, textsizer)



        self.nb.AddPage(fhead, ' File Header ', True)
        self.nb.AddPage(shead, ' Scan Header ', True)
        self.nb.AddPage(self.plotpanel, ' Plot of Selected Arrays ', True)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(panel, 0, LEFT|wx.GROW, 1)
        sizer.Add(self.nb, 1, LEFT|wx.GROW|wx.ALL, 1)
        pack(rightpanel, sizer)

        splitter.SplitVertically(leftpanel, rightpanel, 1)
        self.statusbar = self.CreateStatusBar(2, 0)
        self.statusbar.SetStatusWidths([-1, -1])
        statusbar_fields = [filename, ""]
        for i in range(len(statusbar_fields)):
            self.statusbar.SetStatusText(statusbar_fields[i], i)

        self.set_energy_units()
        csize = self.GetSize()
        bsize = self.GetBestSize()
        if bsize[0] > csize[0]: csize[0] = bsize[0]
        if bsize[1] > csize[1]: csize[1] = bsize[1]
        self.SetSize(csize)
        self.Show()
        self.Raise()
        self.onUpdate(self)

    def set_energy_units(self):
        ix  = self.xarr.GetSelection()
        xname = self.xarr.GetStringSelection()
        rdata = self.curscan.data
        try:
            ncol, npts = rdata.shape
        except:
            self.statusbar.SetStatusText(f"Warning: Could not read data for scan '{self.curscan.title:s}'")

        workgroup = self.workgroup
        if xname.startswith('_index') or ix >= ncol:
            workgroup.xplot = 1.0*np.arange(npts)
        else:
            workgroup.xplot = 1.0*rdata[ix, :]
        eguess =  guess_energy_units(workgroup.xplot)
        if eguess.startswith('eV'):
            self.en_units.SetStringSelection('eV')
        elif eguess.startswith('keV'):
            self.en_units.SetStringSelection('keV')

    def onScanSelect(self, event=None):
        try:
            scan_desc = event.GetString()
            name = [s.strip() for s in scan_desc.split(' | ')][0]
            self.curscan = self.specfile.get_scan(name)
        except:
            return
        slist = list(self.scanlist.GetCheckedStrings())
        if scan_desc not in slist:
            slist.append(scan_desc)
        self.scanlist.SetCheckedStrings(slist)

        self.wid_scantitle.SetLabel("  %s" % self.curscan.title)
        self.wid_scantime.SetLabel(self.curscan.timestring)

        self.title.SetLabel("  %s, scan %s" % (self.path, self.curscan.scan_name))
        self.array_labels = [l.lower() for l in self.curscan.array_labels]
        self.workgroup.array_labels = self.array_labels
        self.workgroup.data = self.curscan.data

        yarr_labels = self.yarr_labels = self.array_labels + ['1.0', '0.0', '']
        xarr_labels = self.xarr_labels = self.array_labels + ['_index']

        xsel = self.xarr.GetStringSelection()
        self.xarr.Clear()
        self.xarr.AppendItems(xarr_labels)
        if xsel in xarr_labels:
            self.xarr.SetStringSelection(xsel)
        else:
            self.xarr.SetSelection(0)

        y1sel = self.yarr1.GetStringSelection()
        self.yarr1.Clear()
        self.yarr1.AppendItems(yarr_labels)
        if y1sel in yarr_labels:
            self.yarr1.SetStringSelection(y1sel)
        else:
            self.yarr1.SetSelection(1)

        y2sel = self.yarr2.GetStringSelection()
        self.yarr2.Clear()
        self.yarr2.AppendItems(yarr_labels)
        if y2sel in yarr_labels:
            self.yarr2.SetStringSelection(y2sel)

        xsel = self.xarr.GetStringSelection()
        self.workgroup.datatype = 'xas' if 'en' in xsel else 'xydata'
        self.datatype.SetStringSelection(self.workgroup.datatype)

        self.scanheader.SetValue('\n'.join(self.curscan.scan_header))
        self.set_energy_units()
        self.onUpdate()

    def onClearMultiColumn(self, event=None):
        self.config['multicol_config'] = {}
        self.message.SetLabel(f" cleared reading of multiple columns")
        self.multi_clear.Disable()
        self.yarr1.Enable()
        self.ypop.Enable()
        self.yop.Enable()
        self.onUpdate()

    def onMultiColumn(self, event=None):
        if 'multicol_config' not in self.config:
            self.config['multicol_config'] = {}

        if len(self.array_labels)  < 1:
            self.array_labels = [l.lower() for l in self.curscan.array_labels]
        self.workgroup.array_labels = self.array_labels
        self.workgroup.data = self.curscan.data
        self.show_subframe('multicol', MultiColumnFrame,
                           config=self.config['multicol_config'],
                           group=self.workgroup,
                           on_ok=self.onMultiColumn_OK)

    def onMultiColumn_OK(self, config, update=True, **kws):
        chans = config.get('channels', [])
        if len(chans) == 0:
            self.config['multicol_config'] = {}
        else:
            self.config['multicol_config'] = config
            self.yarr1.SetSelection(chans[0])
            self.yarr2.SetSelection(config['i0'])
            self.ypop.SetStringSelection('')
            self.yarr1.Disable()
            self.ypop.Disable()
            self.yop.Disable()
            y2 = self.yarr2.GetStringSelection()
            msg = f"  Will import {len(config['channels'])} Y arrays, divided by '{y2}'"
            self.message.SetLabel(msg)
            self.multi_clear.Enable()
        if update:
            self.onUpdate()


    def show_subframe(self, name, frameclass, **opts):
        shown = False
        if name in self.subframes:
            try:
                self.subframes[name].Raise()
                shown = True
            except:
                pass
        if not shown:
            self.subframes[name] = frameclass(self, **opts)
            self.subframes[name].Show()
            self.subframes[name].Raise()

    def set_array_labels(self, arr_labels):
        self.workgroup.array_labels = arr_labels
        yarr_labels = self.yarr_labels = arr_labels + ['1.0', '0.0', '']
        xarr_labels = self.xarr_labels = arr_labels + ['_index']
        def update(wid, choices):
            curstr = wid.GetStringSelection()
            curind = wid.GetSelection()
            wid.SetChoices(choices)
            if curstr in choices:
                wid.SetStringSelection(curstr)
            else:
                wid.SetSelection(curind)
        update(self.xarr,  xarr_labels)
        update(self.yarr1, yarr_labels)
        update(self.yarr2, yarr_labels)
        update(self.yerr_arr, yarr_labels)
        self.onUpdate()

    def onSelAll(self, event=None):
        self.scanlist.SetCheckedStrings(self.scans)

    def onSelNone(self, event=None):
        self.scanlist.SetCheckedStrings([])

    def onOK(self, event=None):
        """ build arrays according to selection """

        scanlist = []
        for s in self.scanlist.GetCheckedStrings():
            words = [s.strip() for s in s.split('|')]
            scanlist.append(words[0])
        if len(scanlist) == 0:
            cancel = Popup(self, """No scans selected.
         Cancel import from this project?""", 'Cancel Import?',
                           style=wx.YES_NO)
            if wx.ID_YES == cancel:
                self.Destroy()
            else:
                return

        self.read_form()
        cout = create_arrays(self.workgroup, **self.config)
        self.config.update(cout)
        conf = self.config
        conf['array_labels'] = self.workgroup.array_labels

        if self.ypop.Enabled:  #not using multicolumn mode
            conf['multicol_config'] = {'channels': [], 'i0': conf['iy2']}

        self.expressions = conf['expressions']

        # generate script to pass back to calling program:
        # read_cmd = "_specfile.get_scan(scan='{scan}')"
        buff = ["{group} = {specfile}.get_scan(scan='{scan}')",
                "{group}.path = '{path}'",
                "{group}.is_frozen = False"]

        for attr in ('datatype', 'plot_xlabel', 'plot_ylabel'):
            val = getattr(self.workgroup, attr)
            buff.append("{group}.%s = '%s'" % (attr, val))

        xexpr = self.expressions['xplot']
        en_units = conf['en_units']
        if en_units.startswith('deg'):
            buff.append(f"mono_dspace = {dspace:.9f}")
            buff.append(f"{{group}}.xplot = PLANCK_HC/(2*mono_dspace*sin(DEG2RAD*({expr:s})))")
        elif en_units.startswith('keV'):
            buff.append(f"{{group}}.xplot = 1000.0*{xexpr:s}")
        else:
            buff.append(f"{{group}}.xplot = {xexpr:s}")

        for aname in ('yplot', 'yerr'):
            expr = self.expressions[aname]
            buff.append(f"{{group}}.{aname} = {expr}")

        dtype = getattr(self.workgroup, 'datatype', 'xytype')
        if dtype == 'xas':
            buff.append("{group}.energy = {group}.xplot"[:])
            buff.append("{group}.mu = {group}.yplot[:]")
            buff.append("sort_xafs({group}, overwrite=True, fix_repeats=True)")
        elif dtype == 'xydata':
            buff.append("{group}.x = {group}.xplot[:]")
            buff.append("{group}.y = {group}.yplot[:]")
            buff.append("{group}.scale = (ptp({group}.yplot)+1.e-15)")
            buff.append("{group}.xshift = 0.0")

        script = "\n".join(buff)

        self.config['array_desc'] = dict(xplot=self.workgroup.plot_xlabel,
                                         yplot=self.workgroup.plot_ylabel,
                                         yerr=self.expressions['yerr'])
        if self.read_ok_cb is not None:
            self.read_ok_cb(script, self.path, scanlist,
                            config=self.config)

        for f in self.subframes.values():
            try:
                f.Destroy()
            except:
                pass
        self.Destroy()

    def onCancel(self, event=None):
        self.workgroup.import_ok = False
        for f in self.subframes.values():
            try:
                f.Destroy()
            except:
                pass
        self.Destroy()

    def onYerrChoice(self, evt=None):
        yerr_choice = evt.GetString()
        self.yerr_arr.Disable()
        self.yerr_val.Disable()
        if 'const' in yerr_choice.lower():
            self.yerr_val.Enable()
        elif 'array' in yerr_choice.lower():
            self.yerr_arr.Enable()
        self.onUpdate()

    def onXSelect(self, evt=None):
        ix  = self.xarr.GetSelection()
        xname = self.xarr.GetStringSelection()

        workgroup = self.workgroup
        rdata = self.curscan.data
        ncol, npts = rdata.shape
        if xname.startswith('_index') or ix >= ncol:
            workgroup.xplot = 1.0*np.arange(npts)
        else:
            workgroup.xplot = 1.0*rdata[ix, :]

        self.monod_val.Disable()
        if self.datatype.GetStringSelection().strip().lower() == 'xydata':
            self.en_units.SetSelection(4)
        else:
            eguess = guess_energy_units(workgroup.xplot)
            if eguess.startswith('keV'):
                self.en_units.SetSelection(1)
            elif eguess.startswith('deg'):
                self.en_units.SetSelection(2)
                self.monod_val.Enable()
            else:
                self.en_units.SetSelection(0)

        self.onUpdate()

    def onEnUnitsSelect(self, evt=None):
        self.monod_val.Enable(self.en_units.GetStringSelection().startswith('deg'))
        self.onUpdate()

    def read_form(self, **kws):
        """return form configuration"""
        datatype = self.datatype.GetStringSelection().strip().lower()
        if datatype == 'xydata':
            self.en_units.SetStringSelection('not energy')

        conf = {'datatype': datatype,
                'ix':  self.xarr.GetSelection(),
                'xarr': self.xarr.GetStringSelection(),
                'en_units': self.en_units.GetStringSelection(),
                'monod': float(self.monod_val.GetValue()),
                'yarr1': self.yarr1.GetStringSelection().strip(),
                'yarr2': self.yarr2.GetStringSelection().strip(),
                'iy1': self.yarr1.GetSelection(),
                'iy2': self.yarr2.GetSelection(),
                'yop': self.yop.GetStringSelection().strip(),
                'ypop': self.ypop.GetStringSelection().strip(),
                'iyerr': self.yerr_arr.GetSelection(),
                'yerr_arr': self.yerr_arr.GetStringSelection(),
                'yerr_op': self.yerr_op.GetStringSelection().lower(),
                'yerr_val': self.yerr_val.GetValue(),
                }
        self.config.update(conf)
        return conf


    def onUpdate(self, value=None, evt=None):
        """column selections changed calc xplot and yplot"""
        workgroup = self.workgroup
        workgroup.data = self.curscan.data
        workgroup.filename = self.curscan.filename

        conf = self.read_form()
        cout = create_arrays(workgroup, **conf)

        self.expression = cout.pop('expressions')
        conf.update(cout)

        if energy_may_need_rebinning(workgroup):
            self.message.SetLabel("Warning: XAS data may need to be rebinned!")

        popts = dict(marker='o', markersize=4, linewidth=1.5,
                     title=Path(workgroup.filename).name,
                     ylabel=workgroup.plot_ylabel,
                     xlabel=workgroup.plot_xlabel,
                     label=workgroup.plot_ylabel)
        try:
            self.plotpanel.plot(workgroup.xplot, workgroup.yplot, **popts)
        except:
            pass


    def plot_messages(self, msg, panel=1):
        self.statusbar.SetStatusText(msg, panel)



'''
This module defines a functions necessary for tomography calculations.

Authors/Modifications:
----------------------
* Margaret Koker, koker@cars.uchicago.edu
'''

import logging
logger = logging.getLogger(__name__)
logger.level = logging.ERROR
logger = logging.getLogger('tomopy.recon')
logger.level = logging.ERROR

import numpy as np
from scipy.optimize import leastsq, minimize

HAS_tomopy = False
try:
    import tomopy
    HAS_tomopy = True
except ImportError:
    pass

TAU = 2.0 * np.pi

# GLOBAL VARIABLES
TOMOPY_ALG = ['gridrec', 'art', 'bart', 'mlem', 'osem', 'ospml_hybrid',
              'ospml_quad', 'pml_hybrid', 'pml_quad', 'sirt' ]

TOMOPY_FILT = ['shepp', 'ramlak', 'butterworth','parzen', 'cosine', 'hann',
               'hamming', 'None']

def ensure_radians(a):
    """ensure angle data is in radians, not degrees,
    converts degrees to radians if a peak-to-peak > 32 or step size > 0.2
    """
    if np.ptp(a) > 32 or np.diff(a).mean() > 0.20:
        a = np.radians(a)
    return a


def reshape_sinogram(A, x, omega):

    ## == INPUTS ==
    ## A              :    array from .get_roimap()
    ## x              :    x array for checking shape of A
    ## omega          :    omega array for checking shape of A
    ##
    ## == RETURNS ==
    ## A              :    A in shape/format needed for tomopy reconstruction
    ## sinogram_order :  flag/argument for tomopy reconstruction (shape dependent)

    if len(x) == len(omega):
        print("Warning: guessing that 2nd axis is omega")
        # Cannot reorder sinogram based on length of positional
        #         arrays when same length. Acceptable orders:
        #         sinogram_order = False : sino = [ 2th   , slice, X ]
        #         sinogram_order = True  : sino = [ slice , 2th  , X ]''')
        if len(A.shape) == 2:
           A = A.reshape(1, A.shape[0], A.shape[1])
        return A, True

    if len(x) < 1 or len(omega) < 1:
        return A,False
    if len(A.shape) != 3:
       if len(A.shape) == 2:
           A = A.reshape(1,A.shape[0],A.shape[1])

    if len(A.shape) == 3:
        if len(x) == A.shape[0]:
             A = np.einsum('kij->ijk', A)
        if len(x) == A.shape[1]:
             A = np.einsum('ikj->ijk', A)
    sinogram_order = len(omega) == A.shape[1]

    return A, sinogram_order

def trim_sinogram(sino, x, omega, pixel_trim=10):
    if len(omega) == sino.shape[-1]:
        omega = omega[pixel_trim:-1*(pixel_trim+1)]
    elif len(x) == sino.shape[-1]:
        x = x[pixel_trim:-1*(pixel_trim+1)]

    sino = sino[:,pixel_trim:-1*(pixel_trim+1)]

    return sino, x, omega

def find_tomo_center(sino, omega, center=None, tol=0.25, blur_weight=2.0,
                     sinogram_order=True):

    """find rotation axis center for a sinogram,
    mixing negative entropy (as tomopy uses) and other focusing scores

    Arguments
    ---------
    sino : ndarray for sinogram
    omega: ndarray of angles in radians
    center: initial value for center [mid-point]
    tol:    fit tolerance for center pixel [0.25]
    sinogram_order: bool for axis order of sinogram

    Returns
    -------
    pixel value for refined center

    Notes
    ------

     The algormithm combines a few focusing scores from Y. Sun, S. Duthaler, and B. Nelson,
     MICROSCOPY RESEARCH AND TECHNIQUE 65:139–149 (2004) (doi: 10.1002/jemt.20118a)

    For a reconstructed image `img` the variance is calculated as

       blur = -((img - img.mean())**2).sum()/img.size

    and is combined with negative-entropy is calculated as
       ioff = (img.max() - img.min())/25.0
       imin = img.min() - ioff
       imax = img.max() + ioff
       hist, _ = np.histogram(img, bins=512, range=[imin, imax])
       hist =  hist/(2*(imax-imin))
       hist[np.where(hist==0)] = 1.e-20
       negent = -np.dot(hist, np.log(hist))


    """
    xmax = sino.shape[0]
    if sinogram_order:
        xmax = sino.shape[2]
    if center is None:
        center = xmax/2.0

    omega = ensure_radians(omega)

    img = tomopy.recon(sino, omega, center,
                       sinogram_order=sinogram_order,
                       algorithm='gridrec', filter_name='shepp')
    img = tomopy.circ_mask(img, axis=0)
    ioff = (img.max() - img.min())/25.0
    imin = img.min() - ioff
    imax = img.max() + ioff

    out = minimize(center_score, center, method='Nelder-Mead', tol=tol,
                   args=(sino, omega, blur_weight, sinogram_order, imin, imax))
    return out.x[0]

def center_score(center, sino, omega, blur_weight=2.0, sinogram_order=True,
                  imin=None, imax=None, verbose=False):
    """Cost function used for the ``find_center`` routine:
    combines a few focusing scores from
    Y. Sun, S. Duthaler, and B. Nelson,
    MICROSCOPY RESEARCH AND TECHNIQUE 65:139–149 (2004)
    (doi: 10.1002/jemt.20118a)

       name   formula in paper   python

        blur (F 10)  -((img - img.mean())**2).sum()/img.size

     negent is calculated as
        hist, _ = np.histogram(img, bins=512, range=[imin, imax])
        hist =  hist/(2*(imax-imin))
        hist[np.where(hist==0)] = 1.e-20
        negent = -np.dot(hist, np.log(hist))

    """
    ns, nang, nx = sino.shape
    if isinstance(center, (list, tuple, np.ndarray)):
        center = center[0]

    img = tomopy.recon(sino, ensure_radians(omega), center,
                       sinogram_order=sinogram_order,
                       algorithm='gridrec', filter_name='shepp')
    img = tomopy.circ_mask(img, axis=0)
    blur = -((img - img.mean())**2).sum()/img.size

    if imin is None or imax is None:
        ioff = (img.max() - img.min())/25.0
        if imin is None:
            imin = img.min() - ioff
        if imax is None:
            imax = img.max() + ioff
    try:
        hist, _ = np.histogram(img, bins=512, range=[imin, imax])
        hist =  hist/(2*(imax-imin))
        hist[np.where(hist==0)] = 1.e-20
        negent = -np.dot(hist, np.log(hist))
    except:
        negent = blur
    if verbose:
        print("Center %.3f %13.5g, %13.5g" % (center, blur, negent))
    return blur*blur_weight + negent


def tomo_reconstruction(sino, omega, algorithm='gridrec',
                        filter_name='shepp', num_iter=1, center=None,
                        refine_center=False, sinogram_order=True):
    '''
    INPUT ->  sino : slice, 2th, x OR 2th, slice, x (with flag sinogram_order=True/False)
    OUTPUT -> tomo : slice, x, y
    '''
    if center is None:
        center = sino.shape[1]/2.

    x = tomopy.init_tomo(sino, sinogram_order)
    nomega = len(omega)
    ns, nth, nx = sino.shape
    if nth > nomega:
        sino = sino[:, :nomega, :]
    romega = ensure_radians(omega)

    if refine_center:
        center = find_tomo_center(sino, romega, center=center,
                                  sinogram_order=sinogram_order)
        print(">> Refine Center done>> ", center, sinogram_order)
    algorithm = algorithm.lower()
    recon_kws = {}
    if algorithm.startswith('gridr'):
        recon_kws['filter_name'] = filter_name
    else:
        recon_kws['num_iter'] = num_iter

    tomo = tomopy.recon(sino, romega, algorithm=algorithm,
                        center=center, sinogram_order=sinogram_order, **recon_kws)
    return center, tomo*(sino.mean()/tomo.mean())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
RIXS data reader for beamline ID26 @ ESRF
=========================================

RIXS stands for Resonant Inelastic X-ray Scattering

**NOTE**: The current implementation is for RIXS files collected at ID26 in
Spec format, that is, before 2019. Currently, ESRF/BLISS format is used at
ID26. Updating this function to the new format should be straightforward by
providing an example of data, but is not yet done/supported.

"""
import os
import time
import numpy as np
from larch.io.specfile_reader import DataSourceSpecH5, _str2rng, _mot2array
from silx.io.dictdump import dicttoh5
from larch.utils.logging import getLogger

_logger = getLogger("io_rixs_id26")
_logger.setLevel("INFO")


def _parse_header(fname):
    """Get parsed header

    Return
    ------
    header : dict
    """
    raise NotImplementedError


def get_rixs_id26(
    fname,
    scans=None,
    sample_name=None,
    mode="rixs",
    mot_axis2="Spec.Energy",
    counter_signal="zap_det_dtc",
    counter_mon="arr_I02sum",
    interp_ene_in=True,
    out_dir=None,
    save_rixs=False,
):
    """Build RIXS map as X,Y,Z 1D arrays

    Parameters
    ----------

    fname : str
        path to the (Spec) data file (TODO: implement BLISS/HDF5)

    scans : str or list of strings or list of ints, optional [None -> all scans in the file]
        list of scans to load (the string is parsed by larch.io.specfile_reader._str2rng)

    sample_name : str, optional ['UNKNOWN_SAMPLE']
        name of the sample measured

    mode : str, optional ['rixs']
        RIXS acqusition mode (affects 'mot_axis2')
            - 'rixs' -> incoming energy scans
            - 'rixs_et' -> emitted energy scans

    mot_axis2 : str ['Spec.Energy']
        name of the counter to use as second axis

    counter_signal : str ['zap_det_dtc']
        name of the counter to use as signal

    counter_mon : str ['arr_I02sum']
        name of the counter to use as incoming beam monitor

    interp_ene_in: bool
        perform interpolation ene_in to the energy step of ene_out [True]

    out_dir : str, optional
        path to save the data [None -> data_dir]

    save_rixs : bool
        if True -> save outdict to disk (in 'out_dir')

    Returns
    -------

    outdict : dict
        {
        '_x': array, energy in
        '_y': array, energy out
        '_z': array, signal
        'mode': str
        'scans': list, scans
        'writer_name': str,
        'writer_version': str,
        'writer_timestamp': str,
        'counter_signal': str, counter_signal,
        'counter_mon': str, counter_mon,
        'mon_axis2': str, mot_axis2,
        'sample_name': str, sample_name,
        'ene_unit': "eV",
        'rixs_header': None,
        'data_dir': str, data_dir,
        'out_dir': str, out_dir,
        'fname_ine': str, full path raw data
        'fname_out': str, full path
        }
    """
    _writer = "get_rixs_id26"
    _writer_version = "1.5.2"  #: used for reading back in RixsData.load_from_h5()
    _writer_timestamp = "{0:04d}-{1:02d}-{2:02d}_{3:02d}{4:02d}".format(
        *time.localtime()
    )
    if sample_name is None:
        sample_name = "SAMPLE_UNKNOWN"

    data_dir = os.path.join(os.sep, *fname.split(os.sep)[1:-1])
    _logger.debug(f"data_dir: {data_dir}")
    if out_dir is None:
        out_dir = data_dir
    ds = DataSourceSpecH5(fname)

    if isinstance(scans, str):
        scans = _str2rng(scans)
    assert isinstance(scans, list), "scans should be a list"

    mode = mode.lower()
    assert mode in ("rixs", "rixs_et"), "RIXS mode not valid"

    _counter = 0
    for scan in scans:
        try:
            ds.set_scan(scan)
            xscan, sig, lab, attrs = ds.get_curve(counter_signal, mon=counter_mon)
        except Exception:
            _logger.error(f"cannot load scan {scan}!")
            continue
        # keV -> eV
        escan = xscan * 1000
        estep = ds.get_motor_position(mot_axis2) * 1000
        if mode == "rixs":
            x = escan
            y = _mot2array(estep, escan)
        if mode == "rixs_et":
            x = _mot2array(estep, escan)
            y = escan
        if _counter == 0:
            xcol = x
            ycol = y
            zcol = sig
        else:
            xcol = np.append(xcol, x)
            ycol = np.append(ycol, y)
            zcol = np.append(zcol, sig)
        _counter += 1
        _logger.info(f"Loaded scan {scan}: {estep:.1f} eV")

    outdict = {
        "_x": xcol,
        "_y": ycol,
        "_z": zcol,
        "mode": mode,
        "scans": scans,
        "writer_name": _writer,
        "writer_version": _writer_version,
        "writer_timestamp": _writer_timestamp,
        "counter_signal": counter_signal,
        "counter_mon": counter_mon,
        "mon_axis2": mot_axis2,
        "sample_name": sample_name,
        "ene_unit": "eV",
        "rixs_header": None,
        "data_dir": data_dir,
        "out_dir": out_dir,
    }

    if save_rixs:
        fnstr = fname.split("/")[-1].split(".")[0]
        fnout = "{0}_rixs.h5".format(fnstr)
        fname_out = os.path.join(out_dir, fnout)
        dicttoh5(outdict, fname_out)
        outdict["fname_in"] = fname
        outdict["fname_out"] = fname_out
        _logger.info("RIXS saved to {0}".format(fnout))

    return outdict


if __name__ == "__main__":
    pass
